"""
Test the feed-archiver representation of an archive of syndication feeeds.
"""

import typing
import unittest
from unittest import mock

from .. import archive
from .. import tests


class FeedarchiverArchiveTests(tests.FeedarchiverTestCase):
    """
    Test the feed-archiver representation of an archive of syndication feeeds.
    """

    # From `./remotes/simple/orig/.../garply%3Fbar=qux%252Fbaz%23corge.rss`
    UPDATE_RETURN_VALUE: typing.Tuple = (["7bd204c6-1655-4c27-aeee-53f933c5395f"], {})

    @mock.patch("feedarchiver.feed.ArchiveFeed")
    def test_feeds_updated(self, mock_feed_class):
        """
        Each feed in the archive configuration is updated.
        """
        mock_update_method = mock_feed_class.return_value.update
        mock_update_method.return_value = self.UPDATE_RETURN_VALUE

        self.archive.load_config()
        mock_feed_class.return_value.config = mock_feed_class.mock_calls[0][2]["config"]
        mock_feed_class.return_value.path = self.feed_path
        mock_feed_class.assert_any_call(
            archive=self.archive,
            config={"remote-url": self.feed_url},
        )
        self.assertEqual(
            mock_feed_class.call_count,
            2,
            "Wrong number of archive feeds instantiated",
        )
        self.archive.archive_feeds[0].url = self.feed_url

        updated_feeds = self.archive.update()
        mock_update_method.assert_called_with()
        self.assertEqual(
            mock_update_method.call_count,
            2,
            "Wrong number of archive feeds updates",
        )
        self.assertIn(
            self.feed_url,
            updated_feeds,
            "RSS feed URL missing from archive updates",
        )
        self.assertEqual(
            updated_feeds[self.feed_url],
            self.UPDATE_RETURN_VALUE,
            "Wrong archive updates RSS feed item",
        )

    @mock.patch("feedarchiver.feed.ArchiveFeed")
    def test_feeds_no_updated_items(self, mock_feed_class):
        """
        Archive handles updates without any updated feed items.
        """
        mock_update_method = mock_feed_class.return_value.update
        mock_update_method.return_value = None
        self.archive.load_config()
        mock_feed_class.return_value.config = mock_feed_class.mock_calls[0][2]["config"]

        updated_feeds = self.archive.update()

        self.assertEqual(
            updated_feeds,
            None,
            "Archive updated feed items not empty",
        )


class FeedarchiverInvalidArchiveTests(unittest.TestCase):
    """
    Test archives with invalid configuration.
    """

    def test_archive_wo_config(self):
        """
        An archive with no configuration raises a helpful error.
        """
        with self.assertRaises(ValueError, msg="Wrong empty config error"):
            archive.Archive(tests.FeedarchiverTestCase.ARCHIVES_PATH / "empty")

    def test_archive_wo_feeds(self):
        """
        An archive with no feeds configured raises a helpful error.
        """
        feed_archive = archive.Archive(
            tests.FeedarchiverTestCase.ARCHIVES_PATH / "empty-feeds",
        )
        with self.assertRaises(ValueError, msg="Wrong empty feeds error"):
            feed_archive.load_config()



"""
Link enclosures about a TV series episode next to the video file as external audio.
"""

import functools
import pathlib
import re
import socket
import logging

try:
    from functools import cached_property  # type: ignore
except ImportError:  # pragma: no cover
    # BBB: Python <3.8 compatibility
    from backports.cached_property import cached_property  # type: ignore

import arrapi
import tenacity

from .. import enclosures

logger = logging.getLogger(__name__)


class SonarrEnclosurePlugin(enclosures.EnclosurePlugin):  # noqa: V102
    """
    Link enclosures about a TV series episode next to the video file as external audio.
    """

    MULTI_EPISODE_RE = re.compile("[Ee& -]")

    url = "http://localhost:8989"
    client = None
    client_get = None

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            (
                socket.error,
                arrapi.exceptions.ConnectionFailure,
            )
        ),
        wait=tenacity.wait_fixed(1),
        stop=tenacity.stop_after_attempt(10),
        reraise=True,
        before_sleep=tenacity.before_sleep_log(logger, logging.DEBUG),
    )
    def load_config(self):
        """
        Pre-process and validate the plugin config prior to linking each enclosure.
        """
        # Validate the configuration
        self.url = self.config.get("url", self.url)
        if not isinstance(self.url, str):  # pragma: no cover
            raise ValueError(f"Sonarr `url` must be a string: {self.url!r}")
        if "api-key" not in self.config:  # pragma: no cover
            raise ValueError("Sonarr plugin configuration must specify `api-key`")
        api_key = self.config["api-key"]
        if not isinstance(api_key, str):  # pragma: no cover
            raise ValueError(f"Sonarr `api_key` must be a string: {api_key!r}")
        self.client = arrapi.SonarrAPI(self.url, api_key)
        self.client_get = self.client._raw._get  # pylint: disable=protected-access

    @cached_property
    def series_by_title(self):
        """
        Request, collate and cache the full list of series titles to share across calls.
        """
        return {series["title"]: series["id"] for series in self.client_get("series")}

    # NOTE: These functions will cache globally, for the life of the process.  This
    # should be fine as `$ feedarchiver update` is expected to be run periodically, such
    # as by `# cron`.
    @functools.lru_cache(maxsize=None)  # pylint: disable=method-cache-max-size-none
    def get_episode_files_seasons(self, series_id):
        """
        Request, correlate and cache episode file DB ids to season and episode numbers.
        """
        episode_files_episodes = {}
        for episode in self.client_get("episode", seriesId=series_id):
            episode_files_episodes.setdefault(episode["episodeFileId"], []).append(
                (episode["seasonNumber"], episode["episodeNumber"])
            )
        return episode_files_episodes

    @functools.lru_cache(maxsize=None)  # pylint: disable=method-cache-max-size-none
    def get_episode_paths(self, series_id):
        """
        Request, collate and cache series episode file paths by season and episode.
        """
        episode_files_episodes = self.get_episode_files_seasons(series_id)
        episode_paths = {}
        for episode_file in self.client_get("episodeFile", seriesId=series_id):
            for season_number, episode_number in episode_files_episodes[
                episode_file["id"]
            ]:
                episode_paths.setdefault(season_number, {})[
                    episode_number
                ] = episode_file["path"]
        return episode_paths

    def __call__(self, enclosure_path, match, *args, **kwargs):
        """
        Lookup the episode and link the enclosure next to the video file.
        """
        series_id, season_number, episode_numbers, stem_append = self.validate_params(
            match,
        )

        # Combine all the parameters to lookup the episode file
        episode_paths = self.get_episode_paths(series_id)
        if season_number not in episode_paths:  # pragma: no cover
            logger.error(
                "Sonarr `season_number` not in series %s episodes: S%s",
                series_id,
                season_number,
            )
        season_episode_paths = episode_paths.get(season_number, {})

        episodes_file_paths = []
        for episode_number in episode_numbers:
            if episode_number in season_episode_paths:
                # Assemble a path next to the episode file
                episode_path = pathlib.Path(season_episode_paths[episode_number])
            else:
                if season_episode_paths:  # pragma: no cover
                    logger.error(
                        "Sonarr `episode_number` not in series %s episodes: S%sE%02d",
                        series_id,
                        season_number,
                        episode_number,
                    )
                # Simulate the episode path
                series = self.client_get(f"series/{series_id}")
                episode_path = pathlib.Path(
                    series["path"],
                    f"Season {season_number:02d}",
                    f"{series['title']} S{season_number}E{episode_number:02d}.mkv",
                )
            episodes_file_paths.append(
                episode_path.with_name(
                    f"{episode_path.stem}{stem_append}" + enclosure_path.suffix,
                ),
            )

        return [str(episode_file_path) for episode_file_path in episodes_file_paths]

    # TODO: Refactor to reduce complexity and improve readability and testibility
    def validate_params(self, match):  # noqa: MC0001
        """
        Combine plugin config and regex match groups, extract and validate parameters.
        """
        # Get API lookup parameters from the config and override with regex match groups
        params = dict(self.config, **match.groupdict())

        # Do any parameter validation that's possible without making API requests
        season_number = params.get("season_number")
        if not season_number:  # pragma: no cover
            raise ValueError(
                f"Sonarr `season_number` missing or empty: {season_number!r}",
            )
        try:
            season_number = int(season_number)
        except ValueError as exc:  # pragma: no cover
            raise ValueError(
                f"Sonarr `season_number` must be an integer: {season_number!r}",
            ) from exc
        episode_numbers_param = params.get("episode_numbers")
        if not episode_numbers_param:  # pragma: no cover
            raise ValueError(
                f"Sonarr `episode_numbers` missing or empty: {episode_numbers_param!r}",
            )
        episode_numbers = []
        for episode_number in self.MULTI_EPISODE_RE.split(episode_numbers_param):
            if not episode_number.strip():
                continue
            try:
                episode_number = int(episode_number.strip())
            except ValueError as exc:  # pragma: no cover
                raise ValueError(
                    f"Sonarr `episode_number` must be an integer: {episode_number!r}",
                ) from exc
            episode_numbers.append(episode_number)
        stem_append = params.get("stem-append", "")
        if not isinstance(stem_append, str):  # pragma: no cover
            raise ValueError(
                f"Sonarr `stem-append` must be a string: {stem_append!r}",
            )

        # Lookup the Sonarr series DB id in order to lookup the episode and file
        series_id = params.get("series_id")
        if not series_id:  # pragma: no cover
            series_title = params.get("series_title")
            if series_title not in self.series_by_title:  # pragma: no cover
                raise ValueError(
                    f"Sonarr `series_title` not in library: {series_title!r}",
                )
            series_id = self.series_by_title[series_title]

        return series_id, season_number, episode_numbers, stem_append



"""
An archive of RSS/Atom syndication feeds.
"""

import os
import pathlib
import urllib.parse
import logging
import tracemalloc
import pdb

import yaml
import httpx
import user_agent
from lxml import etree  # nosec B410
from lxml.html import builder  # nosec B410

from . import utils
from . import feed
from . import enclosures
from .utils import mimetypes

logger = logging.getLogger(__name__)


class Archive:  # pylint: disable=too-many-instance-attributes
    """
    An archive of RSS/Atom syndication feeds.
    """

    INDEX_BASENAME = "index.html"

    FEED_CONFIGS_BASENAME = ".feed-archiver.yml"

    # Initialized when the configuration is loaded prior to update
    global_config = None
    enclosure_plugins = None
    enclosure_fallack_plugins = None
    # The default base URL for assembling absolute URLs
    url = url_split = None
    archive_feeds = None

    def __init__(self, root_dir, recreate=False):
        """
        Instantiate a representation of an archive from a file-system path.
        """
        if utils.PYTHONTRACEMALLOC:  # pragma: no cover
            # Optionally initialize memory profiling
            self.tracemalloc_snapshot = tracemalloc.take_snapshot()

        self.root_path = pathlib.Path(root_dir)
        self.root_stat = os.statvfs(self.root_path)
        self.config_path = self.root_path / self.FEED_CONFIGS_BASENAME
        if not self.config_path.is_file():
            raise ValueError(f"Feeds definition path is not a file: {self.config_path}")

        self.recreate = recreate

        self.client = httpx.Client(follow_redirects=True)
        # Avoid bot detection, real-world `User-Agent` HTTP header values
        self.client.headers.update({"User-Agent": user_agent.generate_user_agent()})

    def load_config(self):
        """
        Read and deserialize the archive feed configs and do necessary pre-processing.
        """
        logger.debug(
            "Retrieving feed configurations: %r",
            str(self.config_path),
        )
        with self.config_path.open(encoding="utf-8") as feeds_opened:
            archive_config = yaml.safe_load(feeds_opened)

        # The first row after the header defines defaults and/or global options
        self.global_config = archive_config["defaults"]
        self.url = self.global_config["base-url"]
        self.url_split = urllib.parse.urlsplit(self.global_config["base-url"])
        (
            self.enclosure_plugins,
            self.enclosure_fallack_plugins,
        ) = enclosures.load_plugins(self, self.global_config)

        feed_configs = archive_config["feeds"]
        if not feed_configs:
            raise ValueError(f"No feeds defined: {str(self.config_path)!r}")

        self.archive_feeds = []
        for feed_config in feed_configs:
            archive_feed = feed.ArchiveFeed(
                archive=self,
                config=feed_config,
            )
            archive_feed.load_config()
            self.archive_feeds.append(archive_feed)

    def response_to_path(self, url_response, url_result=None, request=None):
        """
        Derive the best archive path to represent the given remote URL request response.

        The goals in order of importance are:
        1. Ensure a unique filesystem path for each URL in order to avoid clashes such
           that one download doesn't overwrite another.
        2. Ensure that responses for the archived path from a static site server are as
           correct and well formed as possible, primarily that the extension matches the
           `Content-Type`.
        3. Ensure that derived paths are compatible with most common filesystems.
        4. Try to derive paths that are as human readable as possible given the above.

        Currently this just involves adding or correcting the suffix/extension if it
        doesn't match a `Content-Type` header.
        """
        if request is None:  # pragma: no cover
            if url_response.history:
                request = url_response.history[0].request
            else:
                request = url_response.request
        url_path = self.url_to_path(str(request.url))
        mime_type = None

        # First try to get the MIME type from the response headers
        if url_response.headers.get("Content-Type"):
            mime_type = utils.parse_content_type(
                url_response.headers["Content-Type"],
            )

        # If there's no MIME type in the the response headers, fallback to the element's
        # attribute if available
        if (
            not mime_type
            and hasattr(url_result, "getparent")
            and url_result.getparent().attrib.get("type")
        ):
            mime_type = utils.parse_content_type(
                url_result.getparent().attrib["type"],
            )

        # Fix the suffix/extension if the MIME type doesn't match
        guessed_type, _ = mimetypes.guess_type(url_path.suffix)
        if mime_type and (
            not url_path.suffix
            or (guessed_type is not None and guessed_type != mime_type)
        ):
            # Header doesn't match the extension, guess the most correct extension
            suffix = mimetypes.guess_extension(mime_type, strict=False)
            if suffix:
                url_path = url_path.with_suffix(suffix)

        return self.truncate_path_parts(url_path)

    def url_to_path(self, url):
        """
        Escape the URL to a safe file-system path within the archive.
        """
        split_url = urllib.parse.urlsplit(url)
        # Want a relative path, strip the leading, root slash
        url_relative = split_url.path.lstrip("/")
        # Add explicit index page basename if the URL points to a directory
        if not split_url.path or split_url.path.endswith("/"):
            url_relative += self.INDEX_BASENAME
        # Use `pathlib.PurePosixPath` to split on forward slashes in the URL regardless
        # of what the path separator is for this platform.
        url_path = pathlib.PurePosixPath(url_relative)
        archive_path = (
            pathlib.Path(
                utils.quote_basename(split_url.scheme),
                utils.quote_basename(split_url.netloc),
            )
            / pathlib.Path(
                *(utils.quote_basename(part) for part in url_path.parent.parts),
            )
            # Place the query and fragment from the URL before the extension/suffix in
            # the path
            / (
                utils.quote_basename(
                    split_url._replace(
                        scheme="",
                        netloc="",
                        path=url_path.stem,
                    ).geturl(),
                )
                + url_path.suffix
            )
        )
        # Translate back to platform-native filesystem path separators/slashes
        return self.truncate_path_parts(archive_path)

    def truncate_path_parts(self, path):
        """
        Truncate the basenames of each part of the path to the filesystem maximum.
        """
        truncated_path = pathlib.Path()
        for part in path.parts:
            part_path = pathlib.Path(part)
            truncated_path = truncated_path / (
                part_path.stem[: self.root_stat.f_namemax - len(part_path.suffix)]
                + part_path.suffix
            )
        return truncated_path

    def path_to_url(self, path):
        """
        Un-escape the safe file-system path within the archive to a URL.
        """
        # Also accept strings
        path = pathlib.Path(path)
        # Strip explicit index page basename if the URL should point to a directory
        basename = path.name
        if basename == self.INDEX_BASENAME:
            path = path.parent
        # Strip the archive's path
        archive_path = path.relative_to(self.root_path)
        # Extract any URL query and/or fragment from before the suffix/extension
        stem_split = urllib.parse.urlsplit(urllib.parse.unquote(archive_path.stem))
        # Use `pathlib.PurePosixPath` to split on forward slashes in the URL regardless
        # of what the path separator is for this platform.
        url_parent_path = pathlib.PurePosixPath(
            *[urllib.parse.unquote(part) for part in archive_path.parts[2:-1]]
        )
        # Re-assemble the URL path without the query or fragment from the stem
        url_path = url_parent_path / f"{stem_split.path}{archive_path.suffix}"
        # Make explicit that the URL points to a directory if the path is index HTML
        url_relative = str(url_path)
        if basename == self.INDEX_BASENAME:
            url_relative += "/"
        # Re-assemble the rest of the archive path elements, preserving the query and
        # fragment extracted from the stem.
        split_url = stem_split._replace(
            scheme=urllib.parse.unquote(archive_path.parts[0]),
            netloc=urllib.parse.unquote(archive_path.parts[1]),
            path=url_relative,
        )
        return split_url.geturl()

    def run_feeds_command(self, command, *args, **kwargs):
        """
        Call the sub-command for each feed handling exceptions and aggregating results.
        """
        self.load_config()
        results = {}
        for archive_feed in self.archive_feeds:
            feed_command = getattr(archive_feed, command)
            try:
                feed_results = feed_command(*args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                logger.exception(
                    "Unhandled exception updating feed: %r",
                    archive_feed.url,
                )
                if utils.POST_MORTEM:  # pragma: no cover
                    pdb.post_mortem()
                continue
            if feed_results:
                results[archive_feed.url] = feed_results
            if utils.PYTHONTRACEMALLOC:  # pragma: no cover
                # Optionally compare memory consumption
                self.tracemalloc_snapshot = utils.compare_memory_snapshots(archive_feed)
        if results:
            return results
        return None

    # Sub-commands

    def update(self):
        """
        Request the URL of each feed in the archive and update contents accordingly.
        """
        feeds_results = self.run_feeds_command("update")
        html = builder.HTML(
            builder.HEAD(
                builder.TITLE("Feed Archive Index"),
            ),
            builder.BODY(
                builder.H1("Feed Archive Index"),
                builder.P(
                    "Use the links below to subscribe to the feeds archived here:",
                ),
                builder.UL(
                    *(
                        builder.LI(
                            builder.A(
                                archive_feed.config.get(
                                    "title",
                                    archive_feed.config.get(
                                        "name",
                                        str(archive_feed.path),
                                    ),
                                ),
                                href=str(archive_feed.path),
                            ),
                        )
                        for archive_feed in self.archive_feeds
                    )
                ),
            ),
        )
        logger.info(
            "Writing HTML index: %s",
            self.root_path / "index.html",
        )
        etree.ElementTree(html).write(self.root_path / "index.html", pretty_print=True)
        return feeds_results

    def relink(self):
        """
        Re-link enclosures to the correct locations for the current configuration.
        """
        return self.run_feeds_command("relink")



"""
Test the feed-archiver downloading of enclosures, assets, etc..
"""

import os
import datetime
import pathlib
import urllib

from lxml import etree  # nosec: B410
import respx.models

from .. import utils
from .. import tests


class FeedarchiverDownloadTests(tests.FeedarchiverDownloadsTestCase):
    """
    Test the feed-archiver downloading of enclosures, assets, etc..
    """

    def test_real_requests_disabled(self):
        """
        Confirm that tests will fail if real/external requests are attempted.
        """
        # Put the mocks in place
        self.update_feed(self.archive_feed)
        with self.assertRaises(respx.models.AllMockedAssertionError):
            self.archive.client.get("http://example.com")

    def test_download_file_metadata(self):
        """
        Download file metadata in the archive reflects remote response headers.

        All metadata that can be extracted from the remote response is reflected in the
        file metadata in the archive.
        """
        # Constants specific to this test
        enclosure_archive_path = (
            self.archive.root_path / self.ENCLOSURE_RELATIVE.with_suffix(".mp3")
        )

        # Set the mock file path modification date which is used by the test fixture to
        # set the header on the request mock.
        enclosure_mock_stat = self.ENCLOSURE_MOCK_PATH.stat()
        os.utime(
            self.ENCLOSURE_MOCK_PATH,
            (enclosure_mock_stat.st_atime, self.OLD_DATETIME.timestamp()),
        )

        # Download the enclosure into the archive
        orig_request_mocks = self.mock_remote(self.archive_feed)
        redirect_request_mock = self.client_mock.get(
            self.ENCLOSURE_URL,
        ).respond(
            status_code=302,
            headers={"Location": self.ENCLOSURE_REDIRECT_URL},
        )
        self.archive_feed.update()
        # The archive file's modification time matches.
        self.assertEqual(
            datetime.datetime.fromtimestamp(enclosure_archive_path.stat().st_mtime),
            self.OLD_DATETIME,
            "Archive download modification date doesn't match `Last-Modified` header",
        )

        # The most appropriate file basename is symlinked to the download file
        self.assertEqual(
            redirect_request_mock.call_count,
            1,
            "Wrong number of redirect requests",
        )
        _, target_request_mock = orig_request_mocks[self.ENCLOSURE_REDIRECT_URL]
        self.assertEqual(
            target_request_mock.call_count,
            1,
            "Wrong number of redirect redirect target mock request calls",
        )

        # Test in the absence of the response headers
        self.archive_feed.path.unlink()
        enclosure_archive_path.unlink()
        no_header_request_mock = self.client_mock.get(
            self.ENCLOSURE_URL,
        ).respond(
            content=self.ENCLOSURE_MOCK_PATH.read_bytes(),
        )
        no_header_request_mock.reset()
        self.assert_no_header_download_mtime(
            no_header_request_mock,
            self.archive.root_path / self.ENCLOSURE_RELATIVE.with_suffix(".mp3"),
        )

    def test_downloads(self):  # pylint: disable=too-many-locals
        """
        All files in the archive after update correspond to the fixture.

        This tests for completeness, that nothing in the archive doesn't correspond to a
        request mock in the test fixture and that none of the request mocks in the
        fixture weren't called.
        """
        # Download all feed enclosures and assets
        orig_request_mocks, updated_feeds = self.update_feed(self.archive_feed)
        self.assertGreater(
            len(orig_request_mocks),
            # At least 2 downloads in addition to the feeds themselves
            len(self.archive.archive_feeds) + 1,
            "Too few request mocks registered by test fixture",
        )
        _, download_paths = updated_feeds
        self.assertIsInstance(
            list(download_paths.values())[0],
            str,
            "Wrong download path return type",
        )

        remote_mock_path = self.REMOTES_PATH / self.EXAMPLE_RELATIVE / "orig"
        uncalled_request_mocks = orig_request_mocks.copy()
        # This test is written from the perspective of the archive for completeness.
        # Walk the whole archive so we can make assertions on everything within,
        # including things that shouldn't be there.
        for archive_path, archive_relative in tests.walk_archive(
            self.archive.root_path,
        ):
            with self.subTest(
                msg="Test one feed download",
                archive_relative=str(archive_relative),
            ):
                # Assert that the request mock was called correctly
                download_url, mock_path = self.archive_relative_to_remote_url(
                    archive_relative,
                    remote_mock_path,
                )
                self.assertIn(
                    download_url,
                    uncalled_request_mocks,
                    "No mock registered for download request",
                )
                _, download_request_mock = uncalled_request_mocks.pop(
                    download_url,
                )
                self.assertEqual(
                    download_request_mock.call_count,
                    1,
                    f"Wrong number of requests: {download_url!r}",
                )

                # Assert that the downloaded file in the archive is correct
                self.assertTrue(
                    archive_path.is_file(), "Download is not a file in the archive"
                )
                if archive_path != self.archive_feed.path:
                    self.assertEqual(
                        archive_path.read_bytes(),
                        mock_path.read_bytes(),
                        "Different archived download content from remote",
                    )

        del uncalled_request_mocks[self.ENCLOSURE_REDIRECT_URL]
        for mock_url in list(uncalled_request_mocks.keys()):
            if mock_url.startswith(self.SONARR_URL):  # pragma: no cover
                del uncalled_request_mocks[mock_url]
        self.assertEqual(
            uncalled_request_mocks,
            {},
            # Some request mocks didn't correspond to archive files
            "Archive download missing for request mocks",
        )

        # Assert archive URLs updated
        archive_tree = etree.parse(  # nosec: B320
            self.archive_feed.path.open(),
            parser=utils.XML_PARSER,
        )
        feed_link_split = urllib.parse.urlsplit(
            archive_tree.find("channel").find("link").text,
        )
        feed_link_path = (
            pathlib.PurePosixPath(feed_link_split.path.lstrip("/")) / "index.html"
        )
        self.assertTrue(
            (self.archive.root_path / feed_link_path).is_file(),
            "Wrong feed link index HTML absolute URL in feed",
        )
        feed_href_split = urllib.parse.urlsplit(
            archive_tree.find("channel")
            .find("{http://www.w3.org/2005/Atom}link")
            .attrib["href"]
        )
        feed_href_path = pathlib.PurePosixPath(feed_href_split.path.lstrip("/"))
        self.assertTrue(
            (self.archive_feed.archive.root_path / feed_href_path).is_file(),
            "Wrong feed XML href absolute URL in feed",
        )
        feed_image_split = urllib.parse.urlsplit(
            archive_tree.find("channel")
            .find("{http://www.itunes.com/dtds/podcast-1.0.dtd}image")
            .attrib["href"]
        )
        feed_image_path = pathlib.PurePosixPath(feed_image_split.path.lstrip("/"))
        self.assertTrue(
            (self.archive.root_path / feed_image_path).is_file(),
            "Wrong feed image absolute URL in feed",
        )
        archive_items = archive_tree.find("channel").findall("item")
        archive_item = archive_items[1]
        item_link_split = urllib.parse.urlsplit(archive_item.find("link").text)
        item_link_path = (
            pathlib.PurePosixPath(
                urllib.parse.unquote(item_link_split.path).lstrip("/")
            )
            / "index.html"
        )
        self.assertTrue(
            (self.archive.root_path / item_link_path).is_file(),
            "Wrong item link index HTML absolute URL in feed",
        )
        item_enclosure_split = urllib.parse.urlsplit(
            archive_item.find("enclosure").attrib["url"],
        )
        item_enclosure_path = pathlib.PurePosixPath(
            urllib.parse.unquote(item_enclosure_split.path).lstrip("/"),
        )
        self.assertTrue(
            (self.archive.root_path / item_enclosure_path).is_file(),
            "Wrong item enclosure absolute URL in feed",
        )
        item_media_content_split = urllib.parse.urlsplit(
            archive_item.find("{*}content").attrib["url"],
        )
        item_media_content_path = pathlib.PurePosixPath(
            urllib.parse.unquote(item_media_content_split.path).lstrip("/"),
        )
        self.assertEqual(
            item_media_content_path,
            item_enclosure_path,
            "Wrong item media content absolute URL in feed",
        )
        item_image_split = urllib.parse.urlsplit(
            archive_item.find("{*}image").attrib["href"],
        )
        item_image_path = pathlib.PurePosixPath(
            urllib.parse.unquote(item_image_split.path).lstrip("/"),
        )
        self.assertTrue(
            (self.archive.root_path / item_image_path).is_file(),
            "Wrong item image absolute URL in feed",
        )

        # Assert existing downloads not re-downloaded
        self.archive_feed.update()
        for archive_path, archive_relative in tests.walk_archive(
            self.archive.root_path,
        ):
            with self.subTest(
                msg="Test one feed download",
                archive_relative=str(archive_relative),
            ):
                # Assert that the request mock was called correctly
                download_url, mock_path = self.archive_relative_to_remote_url(
                    archive_relative,
                    remote_mock_path,
                )
                self.assertIn(
                    download_url,
                    orig_request_mocks,
                    "No mock registered for download request",
                )
                _, download_request_mock = orig_request_mocks[download_url]
                if archive_path != self.archive_feed.path:
                    self.assertEqual(
                        download_request_mock.call_count,
                        1,
                        "Request made for already archived download",
                    )



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
# https://kislyuk.github.io/argcomplete/#global-completion
"""
Execute CLI via Python's `-m` option.
"""

import sys

from . import main

sys.exit(main())



"""
The default enclosure plugin, expands a template into the target path.
"""

import pathlib

from .. import utils  # noqa, pylint: disable=unused-import
from .. import enclosures


class TemplateEnclosurePlugin(enclosures.EnclosurePlugin):  # noqa: V102
    """
    The default enclosure plugin, expands a template into the target path.
    """

    # Default a hierarchy under the feed title and item title
    template = str(
        pathlib.Path(
            "Feeds",
            "{utils.quote_sep(feed_parsed.feed.title).strip()}",
            "{utils.quote_sep(item_parsed.title).strip()}{enclosure_path.suffix}",
        )
    )

    def load_config(self):
        """
        Pre-process and validate the plugin config prior to linking each enclosure.
        """
        self.template = self.config.get("template", self.template)
        if not isinstance(self.template, str):  # pragma: no cover
            raise ValueError(
                f"Format `template` for plugin must be a string: {self.template!r}"
            )

    def __call__(
        self,
        *args,
        **kwargs,
    ):  # pylint: disable=too-many-arguments
        """
        Link the feed item enclosure to a target path expanded from a template.
        """
        # Maybe find a better template format/engine, perhaps Jinja2?
        # We need a templating engine that supports a very rich set of operations
        # sufficient, for example, to extract data from `ElementTree` objects.  This
        # means we need to support execution of arbitrary code.  This shouldn't be a
        # problem since anyone that can run `$ feedarchiver` can also run `$ python`.
        # But still, this has a bad code smell.
        # https://python-forum.io/thread-24481.html
        return [
            eval(  # nosec: B307, pylint: disable=eval-used
                f"f{self.template!r}",
                globals(),
                dict(locals(), **kwargs),
            )
        ]



"""
Utility functions or other shared constants and values.

Particularly useful to avoid circular imports.
"""

import os
import functools
import copy
import mimetypes
import urllib.parse
import email
import logging
import tracemalloc

import feedparser

from lxml import etree  # nosec B410

logger = logging.getLogger(__name__)

# Configure the XML parser as securely as possible since we're parsing XML from
# untrusted sources:
# https://lxml.de/FAQ.html#how-do-i-use-lxml-safely-as-a-web-service-endpoint
XML_PARSER = etree.XMLParser(resolve_entities=False)

TRUE_STRS = {"1", "true", "yes", "on"}
DEBUG = (  # noqa: F841
    "DEBUG" in os.environ and os.environ["DEBUG"].strip().lower() in TRUE_STRS
)
POST_MORTEM = (  # noqa: F841
    "POST_MORTEM" in os.environ
    and os.environ["POST_MORTEM"].strip().lower() in TRUE_STRS
)
PYTHONTRACEMALLOC = (
    "PYTHONTRACEMALLOC" in os.environ
    and os.environ["PYTHONTRACEMALLOC"].strip().lower()
)

PRIORITY_TYPES = {
    "application/xml": ".xml",
    "audio/ogg": ".ogg",
    "video/x-matroska": ".mkv",
    # Not actually needed as overrides but found in the wild
    # and not in `/etc/mime.types`
    "application/rss+xml": ".rss",  # `application/x-rss+xml` in `/etc/mime.types`
}


def init(files=None, priority_types=None):
    """
    Fix broken defaults in the Python's `mimetypes` standard library module.

    https://bugs.python.org/issue1043134

    Unfortunately, the defaults in the `mimetypes` module are wrong; `application/xml`
    for example is associated with `.xsl`.  Also unfortunately, `**/mime.types` files
    are also often wrong; `audio/mpeg` for example is associated with `.mpga`, at least
    in Ubuntu's `/etc/mime.types`.  Since both are wrong in different ways, there's no
    avoiding manual intervention.

    For each given priority type, ensure that the extension is returned first.
    Internally, the `mimetypes` module relies on the order in which types are added to
    the registry to decide which extension/suffix is first and thus the default for a
    given MIME type.  As such, for each priority type we manually move the priority
    extension to the front of the list extensions are appended to when they're added.
    Also requires promoting any such types to `strict=True` types if they were
    originally registered as `strict=False`.
    """
    if priority_types is None:  # pragma: no cover
        priority_types = PRIORITY_TYPES

    # Ensure the standard library module has registered all the types first
    mimetypes.init(files=files)
    mimetypes_db = mimetypes._db  # pylint: disable=protected-access
    strict_types_map_inv = mimetypes_db.types_map_inv[True]
    loose_types_map_inv = mimetypes_db.types_map_inv[False]

    # Manually promote the priority extensions to the front of the list
    for priority_type, priority_ext in priority_types.items():
        priority_type = priority_type.lower()
        if priority_type not in strict_types_map_inv:
            # Must re-register as a strict type first
            mimetypes.add_type(priority_type, priority_ext)
        for types_map_inv in (strict_types_map_inv, loose_types_map_inv):
            if priority_type not in types_map_inv:
                continue
            extensions = types_map_inv[priority_type] = list(
                types_map_inv[priority_type],
            )
            if priority_ext not in extensions:  # pragma: no cover
                continue
            extensions.remove(priority_ext)
            extensions.insert(0, priority_ext)


init()


# Abuse URL quoting for paths that are safe across filesystems:
# - *do* quote (IOW, do *not* allow) "/"
# - do *not* quote (IOW, *do* allow) spaces and other special characters found not to
#   cause problems
#
# So far, special characters have been checked in a Samba share as browsed in the
# Windows 10 explorer in order to determine which should be allowed/unquoted.  The `%`
# character works in this test bed but of course it *must* be quoted, otherwise quoting
# and unquoting would not be symmetrical.  A directory with a Unicode character was also
# tested against this environment and found to be working but it doesn't seem possible
# to get `urllib.parse.quote` to leave them unquoted.  Test files were generated in the
# Samba share from the Linux side using the following:
#
#     tmp_path = pathlib.Path("/media/Library/tmp/feed-archiver")
#     [
#         (tmp_path / f"{char_idx}{char}").write_text("")
#         for char_idx, char in enumerate(string.printable)
#         if urllib.parse.quote(char, safe=" ").startswith("%")
#     ]
#
# Please do report any additional cases that cause issues in any other
# common filesystems.
SAFE_CHARS_WIN10_SAMBA = " !#$&'()+,;=@[]^`{}"
QUOTED_SEP = urllib.parse.quote(os.sep, safe="")
QUOTED_ALTSEP = None
if os.altsep is not None:  # pragma: no cover
    QUOTED_ALTSEP = urllib.parse.quote(os.altsep)
quote_basename = functools.partial(urllib.parse.quote, safe=SAFE_CHARS_WIN10_SAMBA)
quote_path = functools.partial(
    urllib.parse.quote,
    safe=f"{SAFE_CHARS_WIN10_SAMBA}{os.sep}{os.altsep}",
)


def quote_sep(string_):  # noqa: V103
    """
    Return the string with all occurrences of path separators, slashes, quoted.

    Useful to sanitize input from feed XML when used in enclosure template plugin string
    formats from adding unintended path parts.
    """
    quoted = string_.replace(os.sep, QUOTED_SEP)
    if os.altsep is not None:  # pragma: no cover
        quoted = quoted.replace(os.altsep, QUOTED_ALTSEP)
    return quoted


def compare_memory_snapshots(parent):  # pragma: no cover
    """
    Compare two traemalloc snapshots and log the results.
    """
    snapshot = tracemalloc.take_snapshot()
    if getattr(parent, "tracemalloc_snapshot", None) is not None:
        stats = snapshot.compare_to(
            parent.tracemalloc_snapshot,
            "lineno",
        )
        logger.debug(
            "Memory consumption changes:\n%s",
            "\n".join(str(stat) for stat in stats[:10]),
        )
    return snapshot


def parse_content_type(content_type):
    """
    Parse an RFC822-style `Content-Type` header.

    Useful to safely extract the MIME type from the charset.
    """
    message = email.message.Message()
    message["Content-Type"] = content_type
    return message.get_params()[0][0]


def copy_empty_items_parent(feed_format, items_parent):
    """
    Create an `etree` copy of the feed items parent without any items.

    Useful for richer parsing of single items at a time.
    """
    items_parent_copy = etree.Element(items_parent.tag, items_parent.attrib)
    for child in items_parent:
        if child.tag == feed_format.ITEM_TAG:
            # Optimization: This is not strictly correct as feed XML may contain
            # non-item elements after feed item elements, either interspersed or at the
            # end.  This is rare, however, in fact I've never seen an instance of it,
            # items are *most* of a feed's elements and use of the items parent other
            # elements in enclosure plugin configurations is rare, so avoid unnecessary
            # iteration until someone reports an issue with this.
            break
        items_parent_copy.append(copy.deepcopy(child))
    return items_parent_copy


# We need to parse the archive and remote feed XML using `etree` because we need to be
# able to modify the XML and write it to the archive, something that `feedparser`
# doesn't provide.  enclosure plugins, however, frequently need the richer parsing
# support that `feedparser` *does* provide, such as parsing dates and times.  That rich
# parsing is only needed in the rare case of new items being added to the archive's
# version of the feed, so only do the rich parsing on a per-item basis.
def parse_item_feed(feed_format, feed_elem, item_elem):
    """
    Reconstruct a "feed" of just one item and return the richly parsed version.
    """
    item_feed_elem = copy_empty_items_parent(feed_format, feed_elem)
    item_feed_elem.append(copy.deepcopy(item_elem))
    return feedparser.parse(etree.tostring(item_feed_elem))



"""
Test the feed-archiver Command-Line Interface.
"""

import sys
import os
import io
import runpy
import subprocess  # nosec B404
import contextlib
import pathlib


import feedarchiver

from .. import tests


class FeedarchiverCLITests(tests.FeedarchiverTestCase):
    """
    Test the feed-archiver Command-Line Interface.
    """

    def test_importable(self):
        """
        The Python package is on `sys.path` and thus importable.
        """
        import_process = subprocess.run(  # nosec B603
            [sys.executable, "-c", "import feedarchiver"],
            check=False,
        )
        self.assertEqual(
            import_process.returncode,
            0,
            "The Python package not importable",
        )

    def get_cli_error_messages(self, args):
        """
        Run the CLI script and return any error messages.
        """
        stderr_file = io.StringIO()
        with self.assertRaises(SystemExit, msg="CLI didn't exit"):
            with contextlib.redirect_stderr(stderr_file):
                feedarchiver.main(args=args)
        return stderr_file.getvalue()

    def test_cli_help(self):
        """
        The command line script is self-docummenting.
        """
        stdout_file = io.StringIO()
        with self.assertRaises(SystemExit, msg="CLI didn't exit"):
            with contextlib.redirect_stdout(stdout_file):
                feedarchiver.main(args=["--help"])
        stdout = stdout_file.getvalue()
        self.assertIn(
            feedarchiver.__doc__.strip(),
            stdout.replace("\n", " "),
            "The console script name missing from --help output",
        )

    def test_cli_subcommand(self):
        """
        The command line supports sub-commands.
        """
        cwd = pathlib.Path.cwd()
        os.chdir(self.archive.root_path)
        try:
            self.assertIsNone(
                feedarchiver.main(args=["update"]),
                "Wrong console script sub-command return value",
            )
        finally:
            os.chdir(cwd)

    def test_cli_option_errors(self):
        """
        The command line script displays useful messages for invalid option values.
        """
        stderr = self.get_cli_error_messages(args=["update", "--non-existent-option"])
        self.assertIn(
            "error: unrecognized arguments: --non-existent-option",
            stderr,
            "Wrong invalid option message",
        )

    def test_cli_dash_m_option(self):
        """
        The package supports execution via Python's `-m` CLI option.
        """
        module_main_process = subprocess.run(  # nosec B603
            [sys.executable, "-m", "feedarchiver", "update"],
            check=False,
            cwd=self.archive.root_path,
        )
        self.assertEqual(
            module_main_process.returncode,
            0,
            "Running via Python's `-m` option exited with non-zero status code",
        )

    def test_cli_module_main(self):
        """
        The package supports execution via Python's `-m` option.
        """
        with self.assertRaises(SystemExit, msg="CLI didn't exit") as exc_context:
            runpy.run_module("feedarchiver")
        self.assertEqual(
            exc_context.exception.code,
            2,
            "Wrong `runpy` exit status code",
        )

    def test_cli_exit_code(self):
        """
        The command line script exits with status code zero if there are no exceptions.
        """
        # Find the location of the `console_scripts` for this test environment
        prefix_path = pathlib.Path(sys.argv[0]).parent
        while not (prefix_path / "bin").is_dir():
            prefix_path = prefix_path.parent
            if prefix_path.parent is list(prefix_path.parents)[-1]:  # pragma: no cover
                raise ValueError(f"Could not find script prefix path: {sys.argv[0]}")

        script_process = subprocess.run(  # nosec B603
            [prefix_path / "bin" / "feed-archiver", "update"],
            check=False,
            cwd=self.archive.root_path,
        )
        self.assertEqual(
            script_process.returncode,
            0,
            "Running the console script exited with non-zero status code",
        )



"""
Handle different feed XML formats.
"""

import typing
import logging

from lxml import etree  # nosec: B410

logger = logging.getLogger(__name__)


def query_xpath(elem, xpath):
    """
    Return the XPath match raising or logging errors if not one.
    """
    xpath_results = elem.xpath(xpath)
    if not xpath_results:  # pragma: no cover
        raise ValueError(f"Matched nothing for {elem.tag!r}: {xpath!r}")
    if len(xpath_results) != 1:  # pragma: no cover
        logger.error(
            "Matched %s results for %r: %r",
            len(xpath_results),
            elem,
            xpath,
        )
    return xpath_results[0]


def all_xpaths_results(elem, xpaths):
    """
    Return the results of all xpaths combined, IOW `OR`.
    """
    return elem.xpath(" | ".join(xpaths))


class FeedFormat:
    """
    An abstract feed XML format, such as RSS or Atom.
    """

    # Constants that define the structural specifics that differ between the formats,
    # mostly tag and attribute names.
    ROOT_TAG = ""
    ITEM_TAG = ""
    ITEM_ID_TAG = ""
    DOWNLOAD_TEXT_TAGS = ["link", "url"]
    DOWNLOAD_ENCLOSURE_TAGS = ["enclosure", "content"]
    DOWNLOAD_ENCLOSURE_EXPR = "@rel='enclosure'"
    DOWNLOAD_ATTR_NAMES = ["href", "url", "src"]

    # XPaths that differ between formats but can't be generalized from the above
    ITEMS_PARENT_XPATH = ""
    SELF_LINK_XPATH = "./*[local-name() = 'link' and @rel = 'self']"

    # Map format-specific sub-classes to their corresponding root tag name.
    # Used to match a parsed feed XML tree to the corresponding format.
    FEED_FORMATS: typing.Dict[typing.Type, typing.Type] = {}

    def __init_subclass__(cls, **kwargs):
        """
        Assemble the XPath components from the format constants.

        Done here so that the resulting XPaths are import-time constants that can be
        overridden by power users, either in their own subclasses or in some future
        configuration option.  The assembled paths can also be printed in logs or CLI
        help so that such power users can use them as a basis for modification.
        """
        super().__init_subclass__(**kwargs)

        cls.ITEMS_XPATH = f"./*[local-name() = '{cls.ITEM_TAG}']"
        cls.ITEM_ID_XPATH = f"./*[local-name() = '{cls.ITEM_ID_TAG}']/text()"

        download_text_expr = cls.assemble_download_text_expr()
        download_attr_step = cls.assemble_download_attr_step()
        download_feed_prefix = (
            f"{cls.ITEMS_PARENT_XPATH}/*[not(local-name() = '{cls.ITEM_TAG}')"
        )
        cls.DOWNLOAD_FEED_URLS_XPATHS = [
            f"{download_feed_prefix} and {download_text_expr}]/text()",
            f"{download_feed_prefix}]//{download_attr_step}",
        ]
        download_enclosure_tags_expr = " or ".join(
            f"local-name()='{download_enclosure_tag}'"
            for download_enclosure_tag in cls.DOWNLOAD_ENCLOSURE_TAGS
        )
        download_enclosure_expr = (
            f"{download_enclosure_tags_expr} or {cls.DOWNLOAD_ENCLOSURE_EXPR}"
        )
        cls.DOWNLOAD_ITEM_ASSET_URLS_XPATHS = [
            f".//*[{download_text_expr}]/text()",
            f".//*[not({download_enclosure_expr})]/{download_attr_step}",
        ]
        cls.DOWNLOAD_ITEM_ENCLOSURE_URLS_XPATHS = [
            f".//*[{download_enclosure_expr}]/{download_attr_step}",
        ]

        # Register this specific feed format by it's root element tag name
        cls.FEED_FORMATS[cls.ROOT_TAG] = cls

        logger.debug(
            "%s XPaths:\n%s",
            cls.__name__,
            "\n".join(
                f"{attr_name}={xpath}"
                for attr_name, xpath in vars(cls).items()
                if attr_name.endswith("_XPATH")
            ),
        )

    def __init__(self, archive_feed):
        """
        Capture a reference to the ArchiveFeed using this feed format.
        """
        self.archive_feed = archive_feed

    @classmethod
    def assemble_download_text_expr(cls):
        """
        Construct the Path predicate expression for text nodes containing URLs.

        Used to combine with other expressions inside different predicates for
        feed-level vs item-level queries.
        """
        elem_expr = " or ".join(
            f"local-name() = '{download_text_tag}'"
            for download_text_tag in cls.DOWNLOAD_TEXT_TAGS
        )
        attr_expr = " or ".join(
            f"@{download_attr_name}" for download_attr_name in cls.DOWNLOAD_ATTR_NAMES
        )
        return f"({elem_expr}) and not({attr_expr})"

    @classmethod
    def assemble_download_attr_step(cls):
        """
        Construct the XPath location step for element attributes containing URLs.
        """
        attr_expr = " or ".join(
            f"local-name()='{download_attr_name}'"
            for download_attr_name in cls.DOWNLOAD_ATTR_NAMES
        )
        return f"@*[{attr_expr}]"

    def get_items_parent(self, feed_root):
        """
        Return the element that contains all feed item elements per the feed format.
        """
        return query_xpath(feed_root, self.ITEMS_PARENT_XPATH)

    def iter_items(self, feed_root):
        """
        Iterate over the feed items given a feed tree root element.
        """
        return self.get_items_parent(feed_root).xpath(self.ITEMS_XPATH)

    def get_item_id(self, item_elem):
        """
        Return the item element value that uniquely identifies it within this feed.
        """
        item_id = query_xpath(item_elem, self.ITEM_ID_XPATH)
        if not isinstance(item_id, str):  # pragma: no cover
            raise ValueError(f"Item ID is not a string: {self.ITEM_ID_XPATH!r}")
        if not item_id.strip():  # pragma: no cover
            raise ValueError(f"Empty feed item ID: {self.ITEM_ID_XPATH!r}")
        return item_id.strip()

    @classmethod
    def from_tree(cls, archive_feed, feed_tree):
        """
        Identify feed format from parsed ElementTree, return corresponding format.

        Be as permissive as possible in identifying the feed format to tolerate poorly
        behaved feeds (e.g. wrong `Content-Type` header, XML namespace, etc.).  Identify
        feeds by the top-level XML tag name (e.g. `<rss>` or `<feed>`).
        """
        feed_root = feed_tree.getroot()
        feed_root_tag = etree.QName(feed_root.tag).localname.lower()
        if feed_root_tag not in cls.FEED_FORMATS:  # pragma: no cover
            raise NotImplementedError(
                f"No feed format handler for {feed_root.tag!r} root element"
            )
        return cls.FEED_FORMATS[feed_root_tag](archive_feed)


class RssFeedFormat(FeedFormat):
    """
    Handle the RSS feed XML format.
    """

    ROOT_TAG = "rss"
    ITEM_TAG = "item"
    ITEM_ID_TAG = "guid"

    ITEMS_PARENT_XPATH = f"/*[local-name() = '{ROOT_TAG}']/*[local-name() = 'channel']"


class AtomFeedFormat(FeedFormat):
    """
    Handle the Atom feed XML format.
    """

    ROOT_TAG = "feed"
    ITEM_TAG = "entry"
    ITEM_ID_TAG = "id"

    ITEMS_PARENT_XPATH = "."



"""
Test the feed-archiver linking of enclosures into media libraries.
"""

import os
import pathlib
import logging

import feedarchiver
from .. import archive
from .. import feed
from .. import tests


class FeedarchiverDownloadTests(tests.FeedarchiverDownloadsTestCase):
    """
    Test the feed-archiver linking of enclosures into media libraries.
    """

    FEED_BASENAME = "Foo Podcast Title"
    ITEM_BASENAME = "El Ni%C3%B1o Episode Title (Qux Series Title 106 & 07)"
    DOWNLOAD_SUFFIX = ".mp3"
    ITEM_DOWNLOAD_BASENAME = f"{ITEM_BASENAME}{DOWNLOAD_SUFFIX}"
    DOWNLOAD_BASENAME = f"download{DOWNLOAD_SUFFIX}"

    def test_basic_feed_item_linking(self):
        """
        By default, enclosures are symlinked to `.../Feed Title/Item Title.ext`.
        """
        feeds_links_path = self.archive_feed.archive.root_path / "Music" / "Podcasts"
        self.assertFalse(
            feeds_links_path.is_dir(),
            "Feed item enclosure symlinks hierarchy exists before updating",
        )

        self.update_feed(self.archive_feed)

        self.assertTrue(
            feeds_links_path.is_dir(),
            "Feed item enclosure symlinks hierarchy does not exist after updating",
        )
        feed_links_path = feeds_links_path / self.FEED_BASENAME
        self.assertTrue(
            feed_links_path.is_dir(),
            "Feed missing from symlinks hierarchy",
        )
        link_path = feed_links_path / self.ITEM_DOWNLOAD_BASENAME
        self.assertTrue(
            link_path.is_symlink(),
            "Item enclosure missing from symlinks hierarchy",
        )
        enclosure_archive_path = (
            self.archive.root_path / self.ENCLOSURE_RELATIVE
        ).with_suffix(".mp3")
        item_enclosures_target = pathlib.Path(
            os.path.relpath(enclosure_archive_path, link_path.parent),
        )
        self.assertEqual(
            os.readlink(link_path),
            str(item_enclosures_target),
            "Item enclosure symlink to wrong target",
        )

    def test_feed_relinking_existing(self):
        """
        The `relink` sub-command cleans up old links and makes new links per the config.
        """
        # Verify initial test fixture
        self.update_feed(self.archive_feed)
        feeds_links_path = self.archive_feed.archive.root_path / "Music" / "Podcasts"
        feed_links_path = feeds_links_path / self.FEED_BASENAME
        link_path = feed_links_path / self.ITEM_DOWNLOAD_BASENAME
        self.assertTrue(
            link_path.is_symlink(),
            "Item enclosure missing from symlinks hierarchy",
        )

        # Change the archive configuration such that downloaded files are symlinked to
        # different locations.
        updated_config_path = (
            self.ARCHIVES_PATH / "relink" / self.archive.config_path.name
        )
        self.archive.config_path.write_text(
            updated_config_path.read_text(),
            encoding="utf-8",
        )

        # Run the `$ feedarchiver relink` sub-command with a configuration that changes
        # the path downloaded files should be linked to.
        feedarchiver.main(
            args=["--archive-dir", str(self.archive_feed.archive.root_path), "relink"],
        )
        self.assertFalse(
            link_path.exists(),
            "Original downloaded file symlink still exists",
        )
        updated_link_path = (
            feed_links_path / f"{self.ITEM_BASENAME} - {self.DOWNLOAD_BASENAME}"
        )
        self.assertTrue(
            updated_link_path.is_symlink(),
            "New downloaded file link is not a symlink",
        )

    def test_feed_relinking_deleted(self):
        """
        The `relink` sub-command tolerates already deleted symlinks.
        """
        self.update_feed(self.archive_feed)
        feeds_links_path = self.archive_feed.archive.root_path / "Music" / "Podcasts"
        feed_links_path = feeds_links_path / self.FEED_BASENAME
        link_path = feed_links_path / self.ITEM_DOWNLOAD_BASENAME
        link_path.unlink()

        # Change the archive configuration such that downloaded files are symlinked to
        # different locations.
        updated_config_path = (
            self.ARCHIVES_PATH / "relink-wo-suffix" / self.archive.config_path.name
        )
        self.archive.config_path.write_text(
            updated_config_path.read_text(),
            encoding="utf-8",
        )
        # Excercise logic to locate feed XML file in archive
        self.feed_path.with_suffix(".txt").write_text("Not actual feed XML")

        # Run the `$ feedarchiver relink` sub-command with a configuration that changes
        # the path downloaded files should be linked to.
        feedarchiver.relink(archive_dir=self.archive_feed.archive.root_path)
        self.assertFalse(
            link_path.exists(),
            "Original downloaded file symlink still exists",
        )
        updated_link_path = (
            feed_links_path / f"{self.ITEM_BASENAME} - {self.DOWNLOAD_BASENAME}"
        )
        self.assertTrue(
            updated_link_path.is_symlink(),
            "New downloaded file link is not a symlink",
        )

    def test_feed_relink_missing_feed(self):
        """
        The `relink` raises a helpful error if the archived feed XML is missing.
        """
        self.update_feed(self.archive_feed)
        self.archive_feed.path.unlink()
        with self.assertLogs(
            archive.logger,
            level=logging.ERROR,
        ) as logged_msgs:
            feedarchiver.relink(archive_dir=self.archive_feed.archive.root_path)
        self.assertEqual(
            len(logged_msgs.records),
            1,
            "Wrong number of download logged archive records",
        )
        self.assertIn(
            "Could not locate feed in archive",
            logged_msgs.records[0].exc_info[1].args[0],
            "Wrong logged record message",
        )

    def test_feed_relink_multiple_feed_files(self):
        """
        The `relink` logs a helpful warning if there are multiple archived feeds.
        """
        self.update_feed(self.archive_feed)
        self.archive_feed.path.with_suffix(".xml").write_text(
            self.archive_feed.path.read_text(),
        )
        self.archive_feed.path.rename(self.archive_feed.path.with_suffix(".xml.rss"))
        with self.assertLogs(
            feed.logger,
            level=logging.WARNING,
        ) as logged_msgs:
            feedarchiver.relink(archive_dir=self.archive_feed.archive.root_path)
        self.assertGreater(
            len(logged_msgs.records),
            0,
            "Missing logged records",
        )
        self.assertIn(
            "Multiple XML files found for feed",
            logged_msgs.records[0].message,
            "Wrong logged record message",
        )


class FeedarchiverRelinkEdgeTests(tests.FeedarchiverTestCase):
    """
    Test `relink` sub-command edge cases.
    """

    EXAMPLE_RELATIVE = pathlib.Path("empty-items")

    def test_feed_relinking_wo_items(self):
        """
        The `relink` sub-command tolerates feeds without items.
        """
        self.update_feed(self.archive_feed)
        relink_results = feedarchiver.relink(
            archive_dir=self.archive_feed.archive.root_path,
        )
        self.assertFalse(
            relink_results,
            "Empty feed returned relink results",
        )



"""
Test updating the archive from the feed URLs.
"""

import os
import datetime
import logging

from lxml import etree  # nosec: B410

from .. import utils
from .. import formats
from .. import feed
from .. import tests


class FeedarchiverFeedTests(tests.FeedarchiverTestCase):
    """
    Test updating the archive from the feed URLs.
    """

    FEED_MOCK_RELATIVE = (
        tests.FeedarchiverTestCase.EXAMPLE_RELATIVE
        / tests.FeedarchiverTestCase.REMOTE_MOCK
        / tests.FeedarchiverTestCase.FEED_ARCHIVE_RELATIVE
    )
    FEED_FORMAT_PARAMS = [
        {
            "feed_format_class": formats.RssFeedFormat,
            "relative_path": FEED_MOCK_RELATIVE,
            "items_parent_tag": "channel",
            "item_tag": "item",
            "item_id": "7bd204c6-1655-4c27-aeee-53f933c5395f",
        },
        {
            "feed_format_class": formats.AtomFeedFormat,
            "relative_path": (
                tests.FeedarchiverTestCase.EXAMPLE_RELATIVE
                / tests.FeedarchiverTestCase.REMOTE_MOCK
                / "https"
                / "waldo.example.com"
                / "feeds"
                / "waldo"
            ),
            "items_parent_tag": (
                f"{{http://www.w3.org/2005/Atom}}{formats.AtomFeedFormat.ROOT_TAG}"
            ),
            "item_tag": "entry",
            "item_id": "urn:uuid:1225c695-cfb8-4ebb-aaaa-80da344efa6a",
        },
    ]

    def test_feed_configs_requested(self):
        """
        Requests are sent for each feed URL in the archive configuration.
        """
        self.assertFalse(
            self.feed_path.exists(),
            "Archive of feed XML exists before updating",
        )
        orig_request_mocks, _ = self.update_feed(self.archive_feed)
        feed_path, get_mock = orig_request_mocks[self.feed_url]
        self.assertEqual(
            get_mock.call_count,
            1,
            "Wrong number of original feed URL requests",
        )
        self.assertTrue(
            self.feed_path.is_file(),
            "Archive of feed XML does not exist after updating",
        )
        with feed_path.open(encoding="utf-8") as remote_opened:
            remote_tree = etree.parse(  # nosec: B320
                remote_opened,
                parser=utils.XML_PARSER,
            )
        remote_items = remote_tree.find("channel").iterchildren("item")
        remote_item_ids = [
            remote_item.find("guid").text for remote_item in remote_items
        ]
        with self.feed_path.open(encoding="utf-8") as archive_opened:
            archive_tree = etree.parse(  # nosec: B320
                archive_opened,
                parser=utils.XML_PARSER,
            )
        archive_items = archive_tree.find("channel").iterchildren("item")
        archive_item_ids = [
            archive_item.find("guid").text for archive_item in archive_items
        ]
        self.assertEqual(
            archive_item_ids,
            remote_item_ids,
            "Archive of feed XML is different from remote",
        )

    def test_feed_unchanged(self):
        """
        Archive feed is untouched if the remote feed XML is unchanged.
        """
        # Confirm initial fixture
        orig_request_mocks, _ = self.update_feed(self.archive_feed)
        feed_path, get_mock = orig_request_mocks[self.feed_url]
        orig_archive_item_elems = tests.get_feed_items(self.feed_path)
        self.assertEqual(
            len(orig_archive_item_elems),
            1,
            "Wrong number of original archived feed items",
        )

        # Feed XML hasn't changed
        self.archive_feed.update()
        self.assertEqual(
            get_mock.call_count,
            2,
            "Wrong number of original feed URL requests",
        )
        with feed_path.open(encoding="utf-8") as remote_opened:
            remote_tree = etree.parse(  # nosec: B320
                remote_opened,
                parser=utils.XML_PARSER,
            )
        remote_items = remote_tree.find("channel").iterchildren("item")
        remote_item_ids = [
            remote_item.find("guid").text for remote_item in remote_items
        ]
        with self.feed_path.open(encoding="utf-8") as archive_opened:
            archive_tree = etree.parse(  # nosec: B320
                archive_opened,
                parser=utils.XML_PARSER,
            )
        archive_items = archive_tree.find("channel").iterchildren("item")
        archive_item_ids = [
            archive_item.find("guid").text for archive_item in archive_items
        ]
        self.assertEqual(
            archive_item_ids,
            remote_item_ids,
            "Archive of feed XML is different from remote",
        )

    def test_feed_added_item(self):
        """
        Items are added to the archive feed XML as the remote feed XML changes.
        """
        # Populate with an archived copy of the original remote feed XML.
        orig_request_mocks, _ = self.update_feed(self.archive_feed)
        _, get_mock = orig_request_mocks[self.feed_url]
        orig_archive_item_elems = tests.get_feed_items(self.feed_path)
        self.assertEqual(
            len(orig_archive_item_elems),
            1,
            "Wrong number of original items in archived feed",
        )

        # Update the archive after the remote feed is updated with a new item added
        get_mock.reset()
        added_item_request_mocks, _ = self.update_feed(
            archive_feed=self.archive_feed,
            remote_mock="added-item",
        )
        _, added_item_get_mock = added_item_request_mocks[self.feed_url]

        # Confirm that the correct request mocks have been used
        self.assertEqual(
            get_mock.call_count,
            1,
            "Wrong number of original feed URL requests",
        )
        self.assertEqual(
            added_item_get_mock.call_count,
            1,
            "Wrong number of added item feed URL requests",
        )

        # The item has been added to the archived feed XML
        added_archive_item_elems = tests.get_feed_items(self.feed_path)
        self.assertEqual(
            len(added_archive_item_elems),
            2,
            "Wrong number of items in archived feed after an item was added",
        )

    def test_feed_removed_item(self):
        """
        Items removed from the remote feed XML are kept in the archive.
        """
        # Populate with an archived copy of the original remote feed XML containing more
        # than one item.
        added_item_request_mocks, _ = self.update_feed(
            archive_feed=self.archive_feed,
        )
        _, added_item_get_mock = added_item_request_mocks[self.feed_url]

        # Update the archive after the remote feed is updated with an item removed
        added_item_get_mock.reset()
        removed_item_request_mocks, _ = self.update_feed(
            archive_feed=self.archive_feed,
            remote_mock="removed-item",
        )
        _, removed_item_get_mock = removed_item_request_mocks[self.feed_url]

        # Confirm that the correct request mocks have been used
        self.assertEqual(
            added_item_get_mock.call_count,
            1,
            "Wrong number of original feed URL requests",
        )
        self.assertEqual(
            removed_item_get_mock.call_count,
            1,
            "Wrong number of removed item feed URL requests",
        )

        # The item has been preserved in the archived feed XML
        removed_archive_item_elems = tests.get_feed_items(self.feed_path)
        self.assertEqual(
            len(removed_archive_item_elems),
            2,
            "Wrong number of items in archived feed after an item was removed",
        )

    def test_feed_reordered_item(self):
        """
        Items are reordered to the archive feed XML as the remote feed XML changes.
        """
        # Populate with a feed containing multiple items
        self.update_feed(
            archive_feed=self.archive_feed,
            remote_mock="added-item",
        )
        # Update the archive after the remote feed is updated with item order changed
        reordered_item_request_mocks, _ = self.update_feed(
            archive_feed=self.archive_feed,
            remote_mock="reordered-item",
        )
        reordered_item_feed_path, _ = reordered_item_request_mocks[self.feed_url]

        remote_tree = etree.parse(  # nosec: B320
            reordered_item_feed_path.open(encoding="utf-8"),
            parser=utils.XML_PARSER,
        )
        remote_items = remote_tree.find("channel").findall("item")
        with self.feed_path.open(encoding="utf-8") as feed_opened:
            archive_tree = etree.parse(  # nosec: B320
                feed_opened,
                parser=utils.XML_PARSER,
            )
        archive_items = archive_tree.find("channel").findall("item")
        self.assertEqual(
            archive_items[0].find("guid").text,
            remote_items[2].find("guid").text,
            "Item added at end of remote feed not at beginning of archive feed",
        )
        self.assertEqual(
            archive_items[1].find("guid").text,
            remote_items[1].find("guid").text,
            "Item original position in remote feed not preserved",
        )
        self.assertEqual(
            archive_items[2].find("guid").text,
            remote_items[0].find("guid").text,
            "Item original position in remote feed not preserved",
        )

    def test_feed_empty(self):
        """
        Updating an empty feed works without any errors.
        """
        # Populate with a feed containing multiple items
        self.update_feed(
            archive_feed=self.archive_feed,
            remote_mock="empty",
        )
        with self.feed_path.open(encoding="utf-8") as feed_opened:
            archive_tree = etree.parse(  # nosec: B320
                feed_opened,
                parser=utils.XML_PARSER,
            )
        archive_items = archive_tree.find("channel").findall("item")
        self.assertEqual(
            archive_items,
            [],
            "Archive contains items for an empty feed",
        )

    def test_feed_formats(self):
        """
        Both RSS and Atom XML feed formats are supported.
        """
        for feed_format_params in self.FEED_FORMAT_PARAMS:
            feed_format_class = feed_format_params["feed_format_class"]
            relative_path = feed_format_params["relative_path"]
            with self.subTest(
                msg="Test one XML feed format",
                feed_format_class=feed_format_class,
            ):
                with (self.REMOTES_PATH / relative_path).open(
                    encoding="utf-8"
                ) as feed_opened:
                    feed_tree = etree.parse(  # nosec: B320
                        feed_opened,
                        parser=utils.XML_PARSER,
                    )
                feed_root = feed_tree.getroot()
                feed_format = feed_format_class(self.archive_feed)

                items_parent = feed_format.get_items_parent(feed_root)
                self.assertEqual(
                    items_parent.tag,
                    feed_format_params["items_parent_tag"],
                    "Wrong feed XML items container element tag name",
                )

                items = list(feed_format.iter_items(feed_root))
                self.assertEqual(
                    len(items),
                    1,
                    "Wrong number of feed XML items",
                )
                self.assertEqual(
                    etree.QName(items[0].tag).localname,
                    feed_format_params["item_tag"],
                    "Wrong feed XML items container element tag name",
                )

                items_id = feed_format.get_item_id(items[0])
                self.assertEqual(
                    items_id,
                    feed_format_params["item_id"],
                    "Wrong feed item unique identifier",
                )

    def test_feed_file_metadata(self):
        """
        Feed file metadata in the archive reflects remote response headers.

        All metadata that can be extracted from the remote response is reflected in the
        file metadata in the archive.
        """
        feed_mock_path = self.REMOTES_PATH / self.FEED_MOCK_RELATIVE
        # Set the mock file path modification date which is used by the test fixture to
        # set the header on the request mock.
        feed_mock_stat = feed_mock_path.stat()
        os.utime(
            feed_mock_path,
            (feed_mock_stat.st_atime, self.OLD_DATETIME.timestamp()),
        )

        # Feed the feed into the archive
        self.update_feed(self.archive_feed)

        # The archive file's modification time matches.
        self.assertEqual(
            datetime.datetime.fromtimestamp(self.feed_path.stat().st_mtime),
            self.OLD_DATETIME,
            "Archive feed modification date doesn't match `Last-Modified` header",
        )

        # Test in the absence of the response headers
        no_header_feed_mock_path = (
            self.REMOTES_PATH
            / self.EXAMPLE_RELATIVE
            / "added-item"
            / tests.FeedarchiverTestCase.FEED_ARCHIVE_RELATIVE
        )
        no_header_request_mock = self.client_mock.get(
            self.feed_url,
        ).respond(
            content=no_header_feed_mock_path.read_bytes(),
        )
        no_header_request_mock.reset()
        self.assert_no_header_download_mtime(
            no_header_request_mock,
            self.feed_path,
        )

    def test_feed_empty_archive_feed(self):
        """
        An empty archive feed file is treated as a first feed request/update.
        """
        # Confirm initial fixture
        orig_request_mocks, _ = self.update_feed(self.archive_feed)
        _, get_mock = orig_request_mocks[self.feed_url]
        orig_archive_item_elems = tests.get_feed_items(self.feed_path)
        self.assertEqual(
            len(orig_archive_item_elems),
            1,
            "Wrong number of original archived feed items",
        )

        # The archive version of the feed was somehow written as an empty file.
        self.feed_path.write_text(
            "{'status': 'Got JSON instead of RSS'",
            encoding="utf-8",
        )
        with self.assertLogs(
            feed.logger,
            level=logging.ERROR,
        ) as logged_msgs:
            self.archive_feed.update()
        self.assertEqual(
            get_mock.call_count,
            2,
            "Wrong number of original feed URL requests",
        )
        self.assertEqual(
            len(logged_msgs.records),
            1,
            "Wrong number of logged feed records",
        )
        self.assertIn(
            "Unhandled exception parsing archive feed",
            logged_msgs.records[0].message,
            "Wrong logged record message",
        )



"""
Plugins for linking feed item enclosures into media libraries.
"""

import re
import pprint

try:
    from importlib.metadata import entry_points  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import entry_points  # type: ignore


def load_plugins(parent, parent_config):
    """
    Pre-process and validate the enclosure plugin configurations.
    """
    archive = getattr(parent, "archive", parent)
    defaults = archive.global_config.get("plugins", {}).get("enclosures", {})
    configs = parent_config.get("enclosures", [])
    if not isinstance(configs, list):  # pragma: no cover
        raise ValueError(
            f"`enclosures` must be a list/array:\n{pprint.pformat(configs)}"
        )
    entrypoints = {ep.name: ep for ep in entry_points()["feedarchiver.enclosures"]}
    plugins = []
    fallack_plugins = []
    for config in configs:
        # Perform any validation as early as possible
        name = config.get("plugin", "default")
        if name not in entrypoints:  # pragma: no cover
            raise ValueError(
                f"enclosure plugin name not registered: {name}",
            )
        if ("match-pattern" in config and "match-string" not in config) or (
            "match-string" in config and "match-pattern" not in config
        ):  # pragma: no cover
            raise ValueError(
                "enclosure plugin config must define either both ``match-pattern`` and"
                " ``match-string`` or neither",
            )
        if "match-pattern" in config:
            config["match-re"] = re.compile(config["match-pattern"])

        # Load, instantiate, and configure the plugin
        plugin_class = entrypoints[name].load()
        # Override global defaults for this plugin with values from this specific
        # instance
        config = dict(
            defaults.get(name, {}),
            **config,
        )
        plugin = plugin_class(parent, config)
        plugin.load_config()

        if plugin.config.get("fallback", False):
            fallack_plugins.append(plugin)
        else:
            plugins.append(plugin)

    return plugins, fallack_plugins


class EnclosurePlugin:
    """
    A plugin for linking feed item enclosures into media libraries.
    """

    def __init__(self, parent, config):
        """
        Instantiate a plugin for linking enclosures into media libraries.
        """
        self.parent = parent
        self.config = config

    def load_config(self):  # pragma: no cover
        """
        Pre-process and validate the plugin config prior to linking each enclosure.
        """
        raise NotImplementedError(
            "Enclosure plugin subclasses must implement the `load_config` method"
        )

    def __call__(self, *args, **kwargs):  # pragma: no cover
        """
        Determine the paths that should be linked to the feed item enclosure.

        Plugin implementations may return a list of paths and let the `ArchiveFeed`
        instance handle the actual linking including appending an numerical index to the
        end of the stem when the link target already exists.  Alternatively, the plugin
        may handle the linking itself and return None.
        """
        raise NotImplementedError(
            "Enclosure plugin subclasses must implement the `__call__` method"
        )



"""
Test the feed-archiver URL escaping for file-system paths.
"""

from .. import tests


class FeedarchiverURLsTests(tests.FeedarchiverTestCase):
    """
    Test the feed-archiver URL escaping for file-system paths.
    """

    def test_url_to_archive_path(self):
        """
        A URL is escaped to a safe file-system path.
        """
        feed_path = self.archive.root_path / self.archive.url_to_path(self.feed_url)
        self.assertEqual(
            feed_path,
            self.feed_path,
            "Wrong safe file-system path for escaped URL",
        )

    def test_archive_path_to_url(self):
        """
        A safe file-system path is un-escaped to a URL.
        """
        feed_url = self.archive.path_to_url(self.feed_path)
        self.assertEqual(
            feed_url,
            self.feed_url,
            "Wrong un-escaped URL for safe file-system path",
        )

    def test_url_roundtrip(self):
        """
        URL escaping to safe filesystem paths is stable and reproducible.
        """
        self.assertEqual(
            self.archive.path_to_url(
                self.archive.root_path / self.archive.url_to_path(self.feed_url),
            ),
            self.feed_url,
            "Different URL after escaping and un-escaping",
        )
        self.assertEqual(
            self.archive.root_path
            / self.archive.url_to_path(
                self.archive.path_to_url(self.feed_path),
            ),
            self.feed_path,
            "Different path after un-escaping and re-escaping",
        )

    def test_url_dir_index(self):
        """
        Trailing slashes in URLs are escaped to an index file: e.g. `index.html`.
        """
        index_path = self.feed_path.parent / "index.html"
        dir_url = self.feed_url.rsplit("/", 1)[0] + "/"
        self.assertEqual(
            self.archive.path_to_url(index_path),
            dir_url,
            "Wrong directory URL for index path",
        )
        self.assertEqual(
            self.archive.root_path / self.archive.url_to_path(dir_url),
            index_path,
            "Wrong index path for directory URL",
        )



"""
An RSS/Atom syndication feed in an archive.
"""

import os
import copy
import re
import urllib
import email.utils
import pathlib
import logging
import pdb

from lxml import etree  # nosec B410

from . import utils
from .utils import mimetypes
from . import formats
from . import enclosures

logger = logging.getLogger(__name__)


class ArchiveFeed:
    """
    An RSS/Atom syndication feed in an archive.
    """

    URL_SCHEME_RE = re.compile(r"^(?P<scheme>[a-zA-Z][a-zA-Z0-9+.-]*):$")
    NAMESPACE = "https://github.com/rpatterson/feed-archiver"

    # Initialized when the configuration is loaded prior to update
    url = None
    enclosure_plugins = None
    enclosure_fallack_plugins = None
    # Initialized on update from the response to the request for the URL from the feed
    # config in order to use response headers to derrive the best path.
    path = None

    def __init__(self, archive, config):
        """
        Instantiate a representation of an archive from a file-system path.
        """
        if utils.PYTHONTRACEMALLOC:  # pragma: no cover
            # Optionally initialize memory profiling
            self.tracemalloc_snapshot = utils.compare_memory_snapshots(archive)

        self.archive = archive
        self.config = config

    def load_config(self):
        """
        Pre-process and validate the feed config prior to running the actual update.
        """
        self.url = self.config["remote-url"]
        self.enclosure_plugins = self.archive.enclosure_plugins[:]
        self.enclosure_fallack_plugins = self.archive.enclosure_fallack_plugins[:]
        enclosure_plugins, enclosure_fallack_plugins = enclosures.load_plugins(
            self,
            self.config,
        )
        self.enclosure_plugins.extend(enclosure_plugins)
        self.enclosure_fallack_plugins.extend(enclosure_fallack_plugins)

    # Sub-commands

    # TODO: Refactor to reduce complexity and improve readability and testibility
    def update(  # noqa: MC0001
        self,
    ):  # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        """
        Request the URL of one feed in the archive and update contents accordingly.
        """
        logger.debug("Requesting feed: %r", self.url)
        remote_response = self.archive.client.get(self.url)
        # Maybe update the extension based on the headers
        self.path = self.archive.root_path / self.archive.response_to_path(
            remote_response,
        )
        remote_tree = self.load_remote_tree(remote_response)
        remote_root = remote_tree.getroot()
        remote_format = formats.FeedFormat.from_tree(self, remote_tree)

        # Assemble the archive version of the feed XML
        download_paths = {}
        archive_tree = self.get_archive_tree(
            remote_tree,
            download_paths,
            remote_format=remote_format,
        )
        archive_root = archive_tree.getroot()
        archived_items_parent = remote_format.get_items_parent(archive_root)

        remote_item_elems = list(remote_format.iter_items(remote_root))
        archive_item_elems = list(remote_format.iter_items(archive_root))
        if (len(remote_item_elems) - len(archive_item_elems)) > 4:
            logger.warning(
                "Many more items in remote than archive: %s > %s",
                len(remote_item_elems),
                len(archive_item_elems),
            )

        logger.info(
            "Updating feed in archive: %r -> %r",
            self.url,
            str(self.path),
        )
        self.update_self(remote_format, archive_root)
        # Iterate through the remote feed to make updates to the archived feed as
        # appropriate.
        archived_item_ids = set()
        updated_items = {}
        # What is the lowest child index for the first item, used to insert new items at
        # the top
        first_item_idx = 0
        for first_item_idx, item_sibling in enumerate(
            archived_items_parent.iterchildren(),
        ):
            if not isinstance(item_sibling.tag, str):
                # Ignore comments or other unusual XML detritus/artifacts
                # `ValueError: Invalid tag name '<cyfunction Comment at 0x...>'`
                continue
            item_sibling_tag = etree.QName(item_sibling.tag).localname
            if item_sibling_tag.lower() == remote_format.ITEM_TAG:
                break
        else:
            first_item_idx += 1
        # Ensure that the order of new feed items is preserved
        remote_item_elems.reverse()
        for remote_item_elem in remote_item_elems:
            logger.debug(
                "Processing remote feed item:\n%s",
                etree.tostring(remote_item_elem).decode(),
            )
            if utils.PYTHONTRACEMALLOC:  # pragma: no cover
                # Optionally compare memory consumption
                self.tracemalloc_snapshot = utils.compare_memory_snapshots(self)

            remote_item_id = remote_format.get_item_id(remote_item_elem)
            if remote_item_id in archived_item_ids:
                # This item was already seen in the archived feed, we don't need to
                # update the archive or search further in the archived feed.
                logger.debug(
                    "Skipping remote feed item already in archived feed: %r",
                    remote_item_id,
                )
                continue
            for archived_item_elem in archive_item_elems:
                archived_item_ids.add(remote_format.get_item_id(archived_item_elem))
                if remote_item_id in archived_item_ids:
                    # Found this item in the archived feed, we don't need to update the
                    # archive and we can stop searching the archived feed for now.
                    # Optimization for the common case where a feed only contains the
                    # most recent items ATM but accrues a lot of items in the archive
                    # over time.
                    logger.debug(
                        "Skipping remote feed item already in archived feed: %r",
                        remote_item_id,
                    )
                    break
            else:
                # The remote item ID was not found in the archived feed, update the feed
                # by adding this remote item.
                logger.info(
                    "Adding feed item to archive: %r -> %r",
                    remote_item_id,
                    str(self.path),
                )

                item_download_paths = self.download_item_enclosures(
                    remote_format,
                    remote_item_elem,
                    remote_item_id,
                )
                if item_download_paths is None:
                    continue
                (
                    item_asset_paths,
                    item_enclosure_paths,
                ) = item_download_paths
                download_paths.update(item_asset_paths)
                download_paths.update(item_enclosure_paths)
                updated_items[remote_item_id] = remote_item_elem

                self.link_item_enclosures(
                    remote_format=remote_format,
                    feed_elem=archived_items_parent,
                    item_elem=remote_item_elem,
                    item_enclosure_paths=item_enclosure_paths,
                )

                archived_items_parent.insert(first_item_idx, remote_item_elem)
                # Pretty format the feed for readability
                etree.indent(archive_tree)
                # Update the archived feed file
                archive_tree.write(str(self.path))

        update_download_metadata(remote_response, self.path)

        if updated_items or download_paths:
            return list(updated_items.keys()), {
                # Return values fit for CLI output
                download_url: str(download_path)
                for download_url, download_path in download_paths.items()
            }
        return None

    # TODO: Refactor to reduce complexity and improve readability and testibility
    def relink(self):  # noqa: MC0001
        """
        Re-link enclosures to the correct locations for this feed.
        """
        # Parse this feed's archive XML
        self.path = self.find_archive_path()
        archive_tree = self.load_archive_tree()
        archive_format = formats.FeedFormat.from_tree(self, archive_tree)
        attr_prefixes = [
            f"{{{self.NAMESPACE}}}enclosure-link-",
            # BBB: old format compatibility
            f"{{{self.NAMESPACE}}}content-link-",
        ]

        # Update the links for each item in this feed
        is_modified = False
        linked_enclosures = {}
        for archive_item_elem in archive_format.iter_items(archive_tree.getroot()):
            item_enclosure_paths = {}
            for url_result in formats.all_xpaths_results(
                archive_item_elem,
                archive_format.DOWNLOAD_ITEM_ENCLOSURE_URLS_XPATHS,
            ):
                for attr, attr_value in list(url_result.getparent().attrib.items()):
                    for attr_prefix in attr_prefixes:
                        if attr.startswith(attr_prefix):
                            break
                    else:
                        continue
                    enclosure_path = pathlib.Path(attr_value)
                    if enclosure_path.is_symlink():
                        logger.info(
                            "Deleting existing enclosure link: %r -> %r",
                            str(enclosure_path),
                            str(os.readlink(enclosure_path)),
                        )
                        enclosure_path.unlink()
                    del url_result.getparent().attrib[attr]
                    is_modified = True
                item_enclosure_paths[url_result] = pathlib.Path(
                    urllib.parse.unquote(
                        urllib.parse.urlsplit(url_result).path.lstrip("/"),
                    ),
                )
            item_enclosures = self.link_item_enclosures(
                remote_format=archive_format,
                feed_elem=archive_format.get_items_parent(archive_tree.getroot()),
                item_elem=archive_item_elem,
                item_enclosure_paths=item_enclosure_paths,
            )
            if item_enclosures:
                is_modified = True
            if is_modified:  # pragma: no cover
                # Pretty format the feed for readability
                etree.indent(archive_tree)
                # Update the archived feed file
                archive_tree.write(str(self.path))
            linked_enclosures.update(
                (
                    url_result,
                    # Make results JSON serializable for CLI stdout
                    [str(enclosure_path) for enclosure_path in enclosure_paths],
                )
                for url_result, enclosure_paths in item_enclosures.items()
            )
        if linked_enclosures:
            return linked_enclosures
        return None

    # Other methods

    def download_item_enclosures(self, remote_format, remote_item_elem, remote_item_id):
        """
        Download all the enclosures from a feed item.
        """
        item_asset_urls = formats.all_xpaths_results(
            remote_item_elem,
            remote_format.DOWNLOAD_ITEM_ASSET_URLS_XPATHS,
        )
        item_enclosure_urls = formats.all_xpaths_results(
            remote_item_elem,
            remote_format.DOWNLOAD_ITEM_ENCLOSURE_URLS_XPATHS,
        )
        try:
            # Download enclosures and assets only for this item.
            item_asset_paths = self.download_urls(
                item_asset_urls,
            )
            item_enclosure_paths = self.download_urls(
                item_enclosure_urls,
            )
        except Exception:  # pylint: disable=broad-except
            logger.exception(
                "Problem downloading item URLs, continuing to next: %r",
                remote_item_id,
            )
            if utils.POST_MORTEM:  # pragma: no cover
                pdb.post_mortem()
            return None
        return item_asset_paths, item_enclosure_paths

    def update_self(self, feed_format, archive_root):
        """
        Update the `<link rel="self" href="..." ...` URL in the archived feed.

        Use the absolute URL so that it can be used as a base URL for other URLs in the
        feed XML.
        """
        archive_base_url_split = urllib.parse.urlsplit(
            self.config.get("base-url") or self.archive.global_config["base-url"]
        )
        base_url_path = pathlib.PurePosixPath(
            archive_base_url_split.path
        ) / urllib.parse.quote(os.path.relpath(self.path, self.archive.root_path))
        base_url_split = archive_base_url_split._replace(path=str(base_url_path))
        for self_link_elem in feed_format.get_items_parent(archive_root).xpath(
            feed_format.SELF_LINK_XPATH
        ):
            self_link_elem.attrib["href"] = base_url_split.geturl()

    def load_remote_tree(self, remote_response):
        """
        Request the feed from the remote URL, parse the XML, and return the tree.

        Also do any pre-processing needed to start updating the archive.
        """
        logger.debug("Parsing remote XML: %r", self.url)
        remote_root = etree.fromstring(  # nosec: B320
            remote_response.content,
            base_url=str(remote_response.url),
            parser=utils.XML_PARSER,
        )
        return etree.ElementTree(remote_root)

    def find_archive_path(self):
        """
        Locate this feed's file in the archive.
        """
        path = self.archive.root_path / self.archive.url_to_path(self.url)
        if not path.exists():
            archive_files = []
            for archive_file in path.parent.glob(f"{path.stem}.*"):
                guessed_type, _ = mimetypes.guess_type(archive_file)
                if guessed_type is not None and (
                    guessed_type.endswith("/xml") or guessed_type.endswith("+xml")
                ):
                    archive_files.append(archive_file)
            if not archive_files:
                raise ValueError(f"Could not locate feed in archive: {self.url}")
            if len(archive_files) > 1:
                logger.warning(
                    "Multiple XML files found for feed, using first: %s\n%s",
                    self.url,
                    "\n  ".join([str(archive_file) for archive_file in archive_files]),
                )
            path = archive_files[0]
        return path

    def load_archive_tree(self):
        """
        Parse the local feed XML in the archive and return the tree.
        """
        archive_tree = None
        if (
            not self.archive.recreate
            and self.path.exists()
            and self.path.read_text().strip()
        ):
            logger.debug("Parsing archive XML: %r", self.url)
            with self.path.open() as feed_archive_opened:
                # Try to parse the local archive version of the feed if possible.  If
                # there are errors parsing it, then treat it as if it's the first time
                # archiving this feed.
                try:
                    archive_tree = etree.parse(  # nosec B320
                        feed_archive_opened,
                        parser=utils.XML_PARSER,
                    )
                except SyntaxError:
                    logger.exception(
                        "Unhandled exception parsing archive feed: %r",
                        self.url,
                    )
                    if utils.POST_MORTEM:  # pragma: no cover
                        pdb.post_mortem()
        return archive_tree

    def get_archive_tree(self, remote_tree, download_paths, remote_format=None):
        """
        Parse the archive feed XML if it exists or initialize an empty archive feed.

        If there is no local feed XML in the archive, such as the first time the feed is
        updated, then initialize the archive tree from the remote tree.

        Also do any pre-processing needed to start updating the archive.
        """
        archive_tree = self.load_archive_tree()
        if archive_tree is not None:
            if remote_format is not None:  # pragma: no cover
                archive_format = formats.FeedFormat.from_tree(self, archive_tree)
                if not isinstance(
                    archive_format,
                    type(remote_format),
                ):  # pragma: no cover
                    raise NotImplementedError(
                        f"Remote feed format, {type(remote_format).__name__!r}, is "
                        "different from archive format, "
                        f"{type(archive_format).__name__!r}."
                    )
            archive_root = archive_tree.getroot()
            archived_items_parent = remote_format.get_items_parent(archive_root)
        else:
            # First time requesting this feed, copy the remote feed, minus the items to
            # the archive and download the feed-level assets.
            logger.info(
                "Initializing feed in archive: %r -> %r",
                self.url,
                str(self.path),
            )
            # Duplicate the remote tree entirely so we can modify the archive's version
            # separately without changing the remote tree and affecting the rest of the
            # update logic.
            archive_tree = copy.deepcopy(remote_tree)
            archive_root = archive_tree.getroot()
            archived_items_parent = remote_format.get_items_parent(archive_root)
            # Remove all feed items from the copied tree.  As items are updated, the
            # modified versions of he items will be added back in during the rest of the
            # update logic.
            for archive_item_elem in remote_format.iter_items(archive_root):
                archived_items_parent.remove(archive_item_elem)

            # Consistent with initial download of the feed, only download assets for the
            # initial version of the feed.
            try:
                archive_download_paths = self.download_urls(
                    formats.all_xpaths_results(
                        archive_root,
                        remote_format.DOWNLOAD_FEED_URLS_XPATHS,
                    ),
                )
            except Exception:  # pragma: no cover, pylint: disable=broad-except
                logger.exception(
                    "Problem downloading feed assets, continuing with items: %s",
                    self.url,
                )
            else:
                download_paths.update(archive_download_paths)

            self.path.parent.mkdir(parents=True, exist_ok=True)
            logger.debug(
                "Writing initialized feed: %r",
                str(self.path),
            )
            etree.indent(archive_tree)
            archive_tree.write(str(self.path))

        return archive_tree

    # TODO: Refactor to reduce complexity and improve readability and testibility
    def download_urls(  # noqa: MC0001
        self,
        url_results,
    ):  # pylint: disable=too-many-branches
        """
        Escape URLs to archive paths, download if new, and update URLs.
        """
        downloaded_paths = {}
        excs = {}
        for url_result in url_results:
            if url_result == self.url:
                # The feed itself is handled in `self.update()`
                continue
            download_path = None
            if utils.PYTHONTRACEMALLOC:  # pragma: no cover
                # Optionally compare memory consumption
                self.tracemalloc_snapshot = utils.compare_memory_snapshots(self)
            if url_result in downloaded_paths:
                logger.debug("Duplicate URL, skipping download: %r", url_result)
                # Proceed below to update the URLs in the duplicate XML element
                download_path = self.archive.root_path / downloaded_paths[url_result]
            else:
                # Download the URL to the escaped local path in the archive
                try:
                    download_path = self.download_url(url_result)
                except Exception as exc:  # pylint: disable=broad-except
                    excs[url_result] = exc
                    logger.exception(
                        "Problem downloading URL, removing from archive: %r",
                        url_result,
                    )
                    if download_path is not None:  # pragma: no cover
                        download_path.unlink()
                    if utils.POST_MORTEM:  # pragma: no cover
                        pdb.post_mortem()
                    continue
                downloaded_paths[url_result] = download_path.relative_to(
                    self.archive.root_path,
                )

            # Update the URL in the feed XML to the relative archive path.
            # Update only after successful download to minimize inconsistent state on
            # errors.
            url_parent = url_result.getparent()
            download_relative = downloaded_paths[url_result]
            if download_relative.name == self.archive.INDEX_BASENAME:
                download_relative = download_relative.parent
            download_url_split = self.archive.url_split._replace(
                # Let pathlib normalize the relative path
                path=str(
                    pathlib.PurePosixPath(self.archive.url_split.path)
                    / urllib.parse.quote(str(download_relative))
                ),
            )
            if url_result.attrname:
                logger.debug(
                    'Updating feed URL: <%s %s="%s"...>',
                    url_result.getparent().tag,
                    url_result.attrname,
                    download_url_split.geturl(),
                )
                # Store the original remote URL in a namespace attribute
                url_parent.attrib[
                    f"{{{self.NAMESPACE}}}attribute-{url_result.attrname}"
                ] = url_result
                # Update the archived URL to the local, relative URL
                url_parent.attrib[url_result.attrname] = download_url_split.geturl()
            else:
                logger.debug(
                    "Updating feed URL: <%s>%s</%s>",
                    url_result.getparent().tag,
                    download_url_split.geturl(),
                    url_result.getparent().tag,
                )
                # Store the original remote URL in a namespace attribute
                url_parent.attrib[f"{{{self.NAMESPACE}}}text"] = url_result
                # Update the URL in the archive XML to the local, relative URL
                url_parent.text = download_url_split.geturl()

        if excs:
            # Ensure the feed item is processed again if any problem occurred with with
            # it's downloads
            raise list(excs.values())[0]
        return downloaded_paths

    def resolve_url(self, url):
        """
        Resolve protocol-relative URLs and workaround other common malformations.
        """
        url_split = urllib.parse.urlsplit(url)
        netloc_scheme_match = self.URL_SCHEME_RE.match(url_split.netloc)
        if url_split.scheme and netloc_scheme_match:
            # Mal-formed repeated scheme URL, e.g.: `href="https://https://example.com"`
            url_split = urllib.parse.urlsplit(
                url_split._replace(scheme="", netloc="")
                ._replace(
                    scheme=netloc_scheme_match.group("scheme"),
                )
                .geturl()
            )
            logger.error(
                "Correcting invalid URL: %r -> %r",
                url,
                url_split.geturl(),
            )
        elif not url_split.scheme and not url_split.netloc:
            # Mal-formed host/netloc-only URL, e.g.: `href="example.com"`
            url_split = urllib.parse.urlsplit("//" + url_split.geturl())
            logger.error(
                "Correcting invalid URL: %r -> %r",
                url,
                url_split.geturl(),
            )
        if url_split.netloc and not url_split.scheme:
            # Protocol/scheme-relative URL, e.g.: `href="//example.com/path"`
            url_split = url_split._replace(
                scheme=urllib.parse.urlsplit(self.url).scheme,
            )
            logger.debug(
                "Resolving protocol-relative URL: %r -> %r",
                url,
                url_split.geturl(),
            )
        return url_split

    def download_url(self, url_result):
        """
        Request a URL and stream the response to the file.
        """
        logger.info("Downloading URL into archive: %r", url_result)
        url_split = self.resolve_url(url_result)
        with self.archive.client.stream(
            "GET",
            url_split.geturl(),
        ) as download_response:
            download_path = self.archive.root_path / self.archive.response_to_path(
                download_response,
                url_result,
            )
            download_relative = download_path.relative_to(self.archive.root_path)
            if download_path.exists():
                logger.debug(
                    "Skipping download already in archive: %r",
                    str(download_relative),
                )
                return download_path
            logger.debug("Writing download into archive: %r", str(download_relative))
            download_path.parent.mkdir(parents=True, exist_ok=True)
            with download_path.open("wb") as download_opened:
                for data in download_response.iter_bytes():
                    download_opened.write(data)

        update_download_metadata(download_response, download_path)

        return download_path

    def link_item_enclosures(
        self,
        remote_format,
        feed_elem,
        item_elem,
        item_enclosure_paths,
    ):
        """
        Link item enclosures into media library hierarchies using plugins.
        """
        link_paths = {}
        if not self.enclosure_plugins and not self.enclosure_fallack_plugins:
            # Avoid unnecessary work when no link plugins are configured, particularly
            # parsing the item with `feedparser`.
            return link_paths

        feed_parsed = utils.parse_item_feed(remote_format, feed_elem, item_elem)
        (item_parsed,) = feed_parsed.entries
        for (
            url_result,
            enclosure_path,
        ) in item_enclosure_paths.items():
            link_idx = 0
            for enclosure_plugin in self.enclosure_plugins:
                for link_path in self.list_plugin_enclosure_links(
                    feed_elem,
                    feed_parsed,
                    item_elem,
                    item_parsed,
                    url_result,
                    enclosure_path,
                    enclosure_plugin,
                ):
                    link_paths.setdefault(url_result, []).append(
                        self.link_plugin_file(
                            url_result,
                            enclosure_path,
                            link_path,
                            link_idx,
                        )
                    )
                    link_idx += 1
            if link_idx > 0:
                # At least one plugin linked the enclosure, skip the fallbacks
                continue
            # Link and fallback configurations for this enclosure
            for enclosure_plugin in self.enclosure_fallack_plugins:
                for link_path in self.list_plugin_enclosure_links(
                    feed_elem,
                    feed_parsed,
                    item_elem,
                    item_parsed,
                    url_result,
                    enclosure_path,
                    enclosure_plugin,
                ):
                    link_paths.setdefault(url_result, []).append(
                        self.link_plugin_file(
                            url_result,
                            enclosure_path,
                            link_path,
                            link_idx,
                        )
                    )
                    link_idx += 1
        return link_paths

    def list_plugin_enclosure_links(
        self,
        feed_elem,
        feed_parsed,
        item_elem,
        item_parsed,
        url_result,
        enclosure_path,
        enclosure_plugin,
    ):  # pylint: disable=too-many-arguments
        """
        Return the links for an individual enclosure and plugin.
        """
        kwargs = {}
        match = None
        if "match-re" in enclosure_plugin.config:
            match_kwargs = locals().copy()
            del match_kwargs["self"]
            match = self.link_item_plugin_match(**match_kwargs)
            if match is None:
                return []
            kwargs.update(match.groupdict())

        # Delegate to the plugin
        logger.debug(
            "Linking item enclosure with %r plugin: %s",
            type(enclosure_plugin),
            str(enclosure_path),
        )
        try:
            link_strs = enclosure_plugin(
                archive_feed=self,
                feed_elem=feed_elem,
                feed_parsed=feed_parsed,
                item_elem=item_elem,
                item_parsed=item_parsed,
                url_result=url_result,
                enclosure_path=enclosure_path,
                match=match,
                **kwargs,
            )
        except Exception:  # pylint: disable=broad-except
            logger.exception(
                "Problem linking item enclosure with %r, continuing to next: %s",
                type(enclosure_plugin),
                str(enclosure_path),
            )
            if utils.POST_MORTEM:  # pragma: no cover
                pdb.post_mortem()
            link_strs = []
        if link_strs is None:  # pragma: no cover
            # Plugin handled any linking itself
            link_strs = []
        if isinstance(link_strs, str):  # pragma: no cover
            link_strs = [link_strs]
        # Filter out duplicate paths but preserve order
        uniq_link_strs = []
        for enclosure_link_str in link_strs:
            if enclosure_link_str not in uniq_link_strs:
                uniq_link_strs.append(enclosure_link_str)
        return [
            self.archive.root_path
            / pathlib.Path(
                utils.quote_path(enclosure_link_str),
            )
            for enclosure_link_str in uniq_link_strs
        ]

    def link_item_plugin_match(self, **kwargs):
        """
        If configured, check for a regular expression match for a feed item enclosure.
        """
        enclosure_plugin = kwargs["enclosure_plugin"]
        kwargs["self"] = self
        try:
            match_string = eval(  # nosec B307, pylint: disable=eval-used
                f"f{enclosure_plugin.config['match-string']!r}",
                globals(),
                kwargs,
            )
        except Exception:  # pylint: disable=broad-except
            logger.exception(
                "Problem expanding `match-string` template for %r: %r",
                type(enclosure_plugin),
                enclosure_plugin.config["match-string"],
            )
            if utils.POST_MORTEM:  # pragma: no cover
                pdb.post_mortem()
            return None
        try:
            match = enclosure_plugin.config["match-re"].match(match_string)
        except Exception:  # pragma: no cover, pylint: disable=broad-except
            logger.exception(
                "Problem matching `match-pattern` for %r: %r",
                type(enclosure_plugin),
                enclosure_plugin.config["match-string"],
            )
            if utils.POST_MORTEM:  # pragma: no cover
                pdb.post_mortem()
            return None
        if match is None:
            logger.debug(
                "The %r plugin `match-pattern` did not match: %r",
                type(enclosure_plugin),
                match_string,
            )
            return None
        return match

    def link_plugin_file(
        self,
        url_result,
        enclosure_archive_relative,
        link_path,
        link_idx,
    ):
        """
        Link an item enclosure to a filesystem path returned by a plugin.
        """
        # Make the link relative
        enclosure_link_target = pathlib.Path(
            os.path.relpath(
                self.archive.root_path / enclosure_archive_relative,
                link_path.parent,
            ),
        )

        # Append numerical index if there are multiple enclosure downloads for this
        # item.
        enclosure_link_stem = link_path.stem
        enclosure_index = 0
        while os.path.lexists(link_path):
            if link_path.is_symlink() and os.readlink(link_path) == str(
                enclosure_link_target
            ):
                logger.debug(
                    "Duplicate item URL, skip enclosure link: %r -> %r",
                    str(link_path),
                    str(enclosure_link_target),
                )
                break
            enclosure_index += 1
            link_path = link_path.with_name(
                enclosure_link_stem[
                    : self.archive.root_stat.f_namemax
                    - (len(link_path.suffix) + len(str(enclosure_index)))
                ]
                + str(enclosure_index)
                + link_path.suffix,
            )
        else:
            logger.info(
                "Linking item enclosure: %r -> %r",
                str(link_path),
                str(enclosure_link_target),
            )
            link_path.parent.mkdir(parents=True, exist_ok=True)
            link_path.symlink_to(enclosure_link_target)

        url_result.getparent().attrib[
            f"{{{self.NAMESPACE}}}enclosure-link-{link_idx}"
        ] = str(link_path)
        return link_path


def update_download_metadata(download_response, download_path):
    """
    Reflect any metdata that can be extracted from the respons in the download file.
    """
    # Set the filesystem modification datetime if the header is provided
    if "Last-Modified" in download_response.headers:
        last_modified = email.utils.parsedate_to_datetime(
            download_response.headers["Last-Modified"],
        )
        feed_stat = download_path.stat()
        os.utime(
            download_path,
            (feed_stat.st_atime, last_modified.timestamp()),
        )
        return last_modified

    return None



# PYTHON_ARGCOMPLETE_OK
"""
Archive RSS/Atom syndication feeds and their enclosures and assets.
"""

import sys
import pathlib
import logging
import argparse
import json
import pdb

import argcomplete

from . import utils

from . import archive

logger = logging.getLogger(__name__)

# Manage version through the VCS CI/CD process
__version__ = None
try:
    from . import version
except ImportError:  # pragma: no cover
    pass
else:  # pragma: no cover
    __version__ = version.version

# Define command line options and arguments
parser = argparse.ArgumentParser(
    description=__doc__.strip(),
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
parser.add_argument(
    "--log-level",
    default=argparse.SUPPRESS,
    # I only wish the `logging` module provided public access to all defined levels
    choices=logging._nameToLevel,  # pylint: disable=protected-access
    help="Select logging verbosity. (default: INFO)",
)
parser.add_argument(
    "--archive-dir",
    "-a",
    help=(
        "the archive root directory into which all feeds, their enclosures and assets "
        "will be downloaded"
    ),
    nargs="?",
    type=pathlib.Path,
    default=pathlib.Path(),
)
# Define CLI sub-commands
subparsers = parser.add_subparsers(
    dest="command",
    required=True,
    help="sub-command",
)


def update(
    archive_dir=parser.get_default("--archive-dir"),
    recreate=parser.get_default("--recreate"),
):  # pylint: disable=missing-function-docstring
    feed_archive = archive.Archive(archive_dir, recreate)
    return feed_archive.update()


update.__doc__ = archive.Archive.update.__doc__
parser_update = subparsers.add_parser(
    "update",
    help=update.__doc__.strip(),  # type: ignore
    description=update.__doc__.strip(),  # type: ignore
)
# Make the function for the sub-command specified in the CLI argument available in the
# argument parser for delegation below.
parser_update.set_defaults(command=update)
parser_update.add_argument(
    "--recreate",
    "-r",
    help="ignore existing feed XML in the archive and rewrite it",
    action="store_true",
)

# Register shell tab completion
argcomplete.autocomplete(parser)


def relink(
    archive_dir=parser.get_default("--archive-dir"),
):  # pylint: disable=missing-function-docstring
    feed_archive = archive.Archive(archive_dir)
    return feed_archive.relink()


relink.__doc__ = archive.Archive.relink.__doc__
parser_relink = subparsers.add_parser(
    "relink",
    help=relink.__doc__.strip(),  # type: ignore
    description=relink.__doc__.strip(),  # type: ignore
)
parser_relink.set_defaults(command=relink)


def config_cli_logging(
    root_level=logging.INFO,
    log_level=parser.get_default("--log-level"),
    **_,
):
    """
    Configure logging CLI usage as early as possible to affect all output.
    """
    # Want just our logger's level, not others', to be controlled by options/environment
    logging.basicConfig(level=root_level)
    # If the CLI option was not specified, fallback to the environment variable
    if log_level is None:  # pragma: no cover
        log_level = "INFO"
        if utils.DEBUG:
            log_level = "DEBUG"
    logger.setLevel(getattr(logging, log_level.strip().upper()))

    # Finer control of external loggers to reduce logger noise or expose information
    # that may be useful to users.
    logging.getLogger("charset_normalizer").setLevel(logging.WARNING)

    return log_level


def main(args=None):  # pylint: disable=missing-function-docstring
    try:
        _main(args=args)
    except Exception:  # pragma: no cover
        if utils.POST_MORTEM:
            pdb.post_mortem()
        raise


def _main(args=None):
    """
    Inner main CLI handler for outer exception handling.
    """
    # Parse CLI options and positional arguments
    parsed_args = parser.parse_args(args=args)
    # Avoid noisy boilerplate, functions meant to handle CLI usage should accept kwargs
    # that match the defined option and argument names.
    cli_kwargs = dict(vars(parsed_args))
    # Remove any meta options and arguments, those used to direct option and argument
    # handling, that shouldn't be passed onto functions meant to handle CLI usage.  More
    # generally, err on the side of options and arguments being kwargs, remove the
    # exceptions.
    del cli_kwargs["command"]
    # Separate the arguments for the sub-command
    prunerr_dests = {
        action.dest for action in parser._actions  # pylint: disable=protected-access
    }
    shared_kwargs = dict(cli_kwargs)
    command_kwargs = {}
    for dest, value in list(shared_kwargs.items()):
        if dest not in prunerr_dests:
            command_kwargs[dest] = value
            del shared_kwargs[dest]

    # Configure logging for CLI usage
    config_cli_logging(**shared_kwargs)
    shared_kwargs.pop("log_level", None)

    # Delegate to the function for the sub-command CLI argument
    logger.debug("Running %r sub-command", parsed_args.command.__name__)
    # Sub-commands may return a result to be pretty printed, or handle output themselves
    # and return nothing.
    result = parsed_args.command(**shared_kwargs, **command_kwargs)
    if result is not None:
        json.dump(result, sys.stdout, indent=2)


main.__doc__ = __doc__



"""
Tests for this feed archiver foundation or template.
"""

import os
import datetime
import pathlib
import tempfile
import shutil
import email.utils
import unittest

from lxml import etree  # nosec: B410

import respx
import requests_mock

from .. import utils
from .. import archive
from ..utils import mimetypes


class FeedarchiverTestCase(
    unittest.TestCase,
):
    """
    Constants and set-up used in all feed-archiver tests.
    """

    maxDiff = None  # noqa: F841

    # A date and time in the past all but guaranteed not to exist naturally in the
    # checkout.
    OLD_DATETIME = datetime.datetime(year=1980, month=11, day=25, hour=8, minute=31)
    # The relative path to the example/sample test data this test will use.
    # Default examples are copied from the Wikipedia page of each RSS/Atom syndication
    # XML format to represent the cleanest, simplest form of feeds.
    EXAMPLE_RELATIVE = pathlib.Path("simple")
    # Relative path to the feed XML content in the checkout test data to be used to mock
    # responses to requests for the remote feed URL.
    REMOTE_MOCK = pathlib.Path("orig")
    # Relative path that corresponds to the remote feed URL within the archive.
    FEED_ARCHIVE_STEM = "garply"
    FEED_ARCHIVE_SUFFIX = ".rss"
    FEED_ARCHIVE_QUERY = "?bar=qux%2Fbaz#corge"
    FEED_ARCHIVE_RELATIVE = pathlib.Path(
        "https",
        "foo-username%3Asecret@grault.example.com",
        "feeds",
        f"{FEED_ARCHIVE_STEM}{utils.quote_basename(FEED_ARCHIVE_QUERY)}"
        f"{FEED_ARCHIVE_SUFFIX}",
    )

    # Test data in the checkout that represents remote feed data
    REMOTES_PATH = pathlib.Path(__file__).parent / "remotes"
    # Test data in the checkout that represents local archived feed data.
    ARCHIVES_PATH = REMOTES_PATH.parent / "archives"

    SONARR_URL = "http://localhost:8989"

    def setUp(self):
        """
        Set up used in all feed-archiver tests.
        """
        super().setUp()

        # Mock HTTP/S requests:
        # https://lundberg.github.io/respx/guide/#mock-httpx
        # TODO: Switch to `assert_all_called=True`
        self.client_mock = respx.MockRouter(assert_all_called=False)
        self.addCleanup(self.client_mock.stop)
        self.client_mock.start()
        # Mock requests made by external libraries, e.g. API client:
        # https://requests-mock.readthedocs.io/en/latest/fixture.html#fixtures
        self.request_mocker = requests_mock.Mocker()
        self.addCleanup(self.request_mocker.stop)
        self.request_mocker.start()

        # Create a temporary directory for mutable test data
        tmp_dir = tempfile.TemporaryDirectory(  # pylint: disable=consider-using-with
            suffix=self.EXAMPLE_RELATIVE.suffix,
            prefix=f"{self.EXAMPLE_RELATIVE.stem}-",
        )
        self.addCleanup(tmp_dir.cleanup)

        # Use the example/sample test data basename to assemble the rest of the
        # filesystem paths used by the tests.
        self.archive_path = self.ARCHIVES_PATH / self.EXAMPLE_RELATIVE

        # Copy the testing example feeds archive
        if pathlib.Path(tmp_dir.name).exists():  # pragma: no cover
            shutil.rmtree(tmp_dir.name)
        shutil.copytree(src=self.archive_path, dst=tmp_dir.name)
        self.archive = archive.Archive(tmp_dir.name)
        # Mock the Sonarr request that is sent when the config is loaded
        self.request_mocker.get(
            f"{self.SONARR_URL}/api/v3/system/status?apikey=secret",
            json={"version": "3.0.6.1342"},
        )
        self.archive.load_config()
        self.archive_feed = self.archive.archive_feeds[0]
        self.feed_url = self.archive_feed.url
        self.feed_path = self.archive.root_path / self.FEED_ARCHIVE_RELATIVE

    def mock_remote(self, archive_feed, remote_mock=None):
        """
        Mock the request responses with the mock dir.

        The relative paths in the mock dir are un-escaped to URLs and used to create the
        request mocks for those URLs.
        """
        if remote_mock is None:
            remote_mock = self.REMOTE_MOCK

        request_mocks = {}
        remote_mock_path = self.REMOTES_PATH / self.EXAMPLE_RELATIVE / remote_mock
        for mock_path, _ in walk_archive(remote_mock_path):
            mock_stat = mock_path.stat()
            mock_bytes = mock_path.read_bytes()
            mock_headers = {
                "Last-Modified": email.utils.formatdate(
                    timeval=mock_stat.st_mtime,
                    usegmt=True,
                ),
                "Content-Length": str(len(mock_bytes)),
            }
            mock_type, _ = mimetypes.guess_type(mock_path.name)
            if mock_type:
                mock_headers["Content-Type"] = mock_type
            mock_url = archive_feed.archive.path_to_url(
                archive_feed.archive.root_path / mock_path.relative_to(remote_mock_path)
            )
            if mock_url.startswith(self.SONARR_URL):
                # Mock Servarr requests using the same library as API client
                mock_response = self.request_mocker.get(
                    mock_url,
                    # Ensure the download response includes `Last-Modified`
                    headers=mock_headers,
                    content=mock_bytes,
                )
            else:
                mock_response = self.client_mock.get(
                    mock_url,
                ).respond(
                    # Ensure the download response includes `Last-Modified`
                    headers=mock_headers,
                    content=mock_bytes,
                )
            request_mocks[mock_url] = (mock_path, mock_response)
        return request_mocks

    def update_feed(self, archive_feed, remote_mock=None):
        """
        Mock the request responses with the mock dir and update the archive.
        """
        request_mocks = self.mock_remote(archive_feed, remote_mock)
        updated_feeds = archive_feed.update()
        return request_mocks, updated_feeds

    def assert_no_header_download_mtime(self, no_header_request_mock, download_path):
        """
        Assert that download responses without headers are handled gracefully.
        """
        self.archive_feed.update()
        self.assertEqual(
            no_header_request_mock.call_count,
            1,
            "Wrong number of feed URL requests without metadata headers",
        )
        self.assertNotEqual(
            datetime.datetime.fromtimestamp(download_path.stat().st_mtime),
            self.OLD_DATETIME,
            "Archive feed modification date not current",
        )


class FeedarchiverDownloadsTestCase(FeedarchiverTestCase):
    """
    Test against a rich, full-featured feed including enclosures.
    """

    EXAMPLE_RELATIVE = pathlib.Path("downloads")

    # Constants specific to this test suite
    FEED_ARCHIVE_RELATIVE = pathlib.Path(
        "https",
        "foo.example.com",
        "podcast",
        "feed.rss",
    )
    ITEM_SLUG_PREFIX = "el-ni%C3%B1o"
    ITEM_SLUG = f"{ITEM_SLUG_PREFIX}-episode-title"
    ENCLOSURE_URL = f"https://foo.example.com/podcast/episodes/{ITEM_SLUG}/download"
    ENCLOSURE_RELATIVE = pathlib.Path(
        "https",
        "foo.example.com",
        "podcast",
        "episodes",
        utils.quote_basename(ITEM_SLUG),
        "download",
    )
    ENCLOSURE_MOCK_PATH = (
        FeedarchiverTestCase.REMOTES_PATH
        / EXAMPLE_RELATIVE
        / FeedarchiverTestCase.REMOTE_MOCK
        / ENCLOSURE_RELATIVE
    )
    ENCLOSURE_REDIRECT_URL = f"https://bar.example.com/media/{ITEM_SLUG_PREFIX}.mp3"

    def archive_relative_to_remote_url(self, archive_relative, remote_mock_path):
        """
        Return the remote URL for the archive path adjusted for the mocks.
        """
        mock_path = remote_mock_path / archive_relative
        remote_url_path = self.archive.root_path / archive_relative
        if archive_relative == self.ENCLOSURE_RELATIVE.with_suffix(".mp3"):
            # Adjust for the case where the remote URL is missing the
            # suffix/extension
            remote_url_path = remote_url_path.with_suffix("")
            mock_path = mock_path.with_suffix("")
        return self.archive.path_to_url(remote_url_path), mock_path


def get_feed_items(feed_path):
    """
    Map item ID to item element for all the items in the given feed XML.
    """
    with feed_path.open() as feed_opened:
        tree = etree.parse(feed_opened, parser=utils.XML_PARSER)  # nosec: B320
    channel = tree.getroot().find("channel")
    return {
        item_elem.find("guid").text: item_elem for item_elem in channel.iter("item")
    }


def walk_archive(archive_root_path):
    """
    Walk the given path like `os.path` but excluding what we want to ignore for tests.
    """
    for root, _, files in os.walk(archive_root_path, followlinks=True):
        root_path = pathlib.Path(root)
        root_relative = root_path.relative_to(archive_root_path)
        # Directories we don't want to "descend" into
        if {
            archive_root_part
            for archive_root_part in pathlib.Path(root).parts
            if archive_root_part.endswith("~")
        } or (
            root_relative.parts
            and root_relative.parts[0]
            in {
                "Feeds",
                "Music",
                "Videos",
            }
        ):  # pragma: no cover
            continue
        for archive_basename in files:
            if (
                archive_basename.endswith("~")
                or archive_basename == archive.Archive.FEED_CONFIGS_BASENAME
            ):  # pragma: no cover
                continue
            archive_path = root_path / archive_basename
            archive_relative = archive_path.relative_to(archive_root_path)
            yield archive_path, archive_relative
